__all__ = ["configure_pytorch", "show_cuda_info", "show_cudnn_info", "setup_runner"]

import logging
from typing import Union

from torch.backends import cuda, cudnn

from gravitorch.runners.base import BaseRunner
from gravitorch.utils.format import str_target_object, to_pretty_dict_str

logger = logging.getLogger(__name__)


def setup_runner(runner: Union[BaseRunner, dict]) -> BaseRunner:
    r"""Sets up the runner.

    The runner is instantiated from its configuration by using the
    ``BaseRunner`` factory function.

    Args:
        runner (``BaseRunner`` or dict): Specifies the runner or its
            configuration.

    Returns:
        ``BaseRunner``: The instantiated runner.
    """
    if isinstance(runner, dict):
        logger.debug(f"Initializing a runner from its configuration... {str_target_object(runner)}")
        runner = BaseRunner.factory(**runner)
    return runner


def configure_pytorch(cudnn_benchmark: bool = False, cudnn_deterministic: bool = False) -> None:
    r"""Configure some PyTorch options.

    Args:
        cudnn_benchmark (bool, optional): If True, causes cuDNN to
            benchmark multiple convolution algorithms and select the
            fastest. You should not set this option to true if you
            want to reproduce result. See
            https://pytorch.org/docs/stable/notes/randomness.html
            for more information. Default: False
        cudnn_deterministic (bool, optional): If True, causes cuDNN
            to only use deterministic convolution algorithms.
            Default: ``False``
    """
    cudnn.benchmark = cudnn_benchmark
    cudnn.deterministic = cudnn_deterministic
    show_cuda_info()
    show_cudnn_info()


def show_cuda_info() -> None:
    r"""Shows information about the CUDA backend."""
    prefix = "torch.backends.cuda.matmul"
    info = {
        f"{prefix}.allow_fp16_reduced_precision_reduction": (
            cuda.matmul.allow_fp16_reduced_precision_reduction
        ),
        f"{prefix}.allow_tf32": cuda.matmul.allow_tf32,
        f"{prefix}.is_built": cuda.is_built(),
    }
    logger.info(
        f"CUDA backend information:\n{to_pretty_dict_str(info, sorted_keys=True, indent=2)}\n"
    )


def show_cudnn_info() -> None:
    r"""Shows information about the cuDNN backend."""
    info = {
        "torch.backends.cudnn.allow_tf32": cudnn.allow_tf32,
        "torch.backends.cudnn.benchmark": cudnn.benchmark,
        "torch.backends.cudnn.deterministic": cudnn.deterministic,
        "torch.backends.cudnn.enabled": cudnn.enabled,
        "torch.backends.cudnn.is_available": cudnn.is_available(),
        "torch.backends.cudnn.version": cudnn.version(),
    }
    logger.info(
        f"cuDNN backend information:\n{to_pretty_dict_str(info, sorted_keys=True, indent=2)}\n"
    )



__all__ = [
    "BaseIterDataPipeCreator",
    "EpochRandomIterDataPipeCreator",
    "SequentialCreatorIterDataPipeCreator",
    "SequentialIterDataPipeCreator",
    "create_sequential_iter_datapipe",
    "setup_iter_datapipe_creator",
]

from gravitorch.creators.datapipe.base import (
    BaseIterDataPipeCreator,
    setup_iter_datapipe_creator,
)
from gravitorch.creators.datapipe.random import EpochRandomIterDataPipeCreator
from gravitorch.creators.datapipe.sequential import (
    SequentialCreatorIterDataPipeCreator,
    SequentialIterDataPipeCreator,
    create_sequential_iter_datapipe,
)



from __future__ import annotations

__all__ = ["BaseEngine"]

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from objectory import AbstractFactory
from torch.nn import Module
from torch.optim import Optimizer

from gravitorch.lr_schedulers import LRSchedulerType
from gravitorch.utils.artifacts import BaseArtifact
from gravitorch.utils.events import BaseEventHandler
from gravitorch.utils.exp_trackers.steps import Step
from gravitorch.utils.history import BaseHistory

if TYPE_CHECKING:
    from gravitorch.datasources import BaseDataSource


class BaseEngine(ABC, metaclass=AbstractFactory):
    r"""Defines the base engine.

    This is an experimental API and the engine design may change in the
    future.
    """

    @property
    @abstractmethod
    def data_source(self) -> BaseDataSource:
        r"""``BaseDataSource``: The data source object associated to
        the engine."""

    @property
    @abstractmethod
    def epoch(self) -> int:
        r"""int: The epoch value. The epoch is 0-based, i.e. the first
        epoch is 0. The value ``-1`` is used to indicate the training
        has not started.
        """

    @property
    @abstractmethod
    def iteration(self) -> int:
        r"""int: The iteration value. The iteration is 0-based, i.e.
        the first iteration is 0. The value ``-1`` is used to indicate
        the training has not started.
        """

    @property
    @abstractmethod
    def lr_scheduler(self) -> LRSchedulerType | None:
        r"""``LRSchedulerType`` or ``None``: The learning rate (LR) scheduler
        if it is defined, otherwise ``None``."""

    @property
    @abstractmethod
    def max_epochs(self) -> int:
        r"""int: The maximum number of training epochs."""

    @property
    @abstractmethod
    def model(self) -> Module:
        r"""``torch.nn.Module``: The model to train and/or evaluate."""

    @property
    @abstractmethod
    def optimizer(self) -> Optimizer | None:
        r"""``torch.nn.Optimizer`` or ``None``: The optimizer to train the
        model.

        It can be ``None`` if the model is not trained.
        """

    @property
    @abstractmethod
    def random_seed(self) -> int:
        r"""int: The random seed."""

    @property
    @abstractmethod
    def should_terminate(self) -> bool:
        r"""bool: Flag to indicate if this engine should terminate training
        at the end of the current epoch.

        If ``True``, the engine should terminate at the end of the
        current epoch. Use the ``terminate()`` method to set this flag
        to ``True``.
        """

    @abstractmethod
    def add_event_handler(self, event: str, event_handler: BaseEventHandler) -> None:
        r"""Adds an event handler to an event.

        The event handler will be called everytime the event happens.

        Args:
            event (str): Specifies the event to attach the event
                handler.
            event_handler (``BaseEventHandler``): Specifies the event
                handler to attach to the event.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> from gravitorch.utils.events import VanillaEventHandler
            >>> engine: BaseEngine = ...  # Create an engine
            >>> def hello_handler():
            ...     print('Hello!')
            >>> engine.add_event_handler('my_event', VanillaEventHandler(hello_handler))
        """

    @abstractmethod
    def add_history(self, history: BaseHistory, key: str | None = None) -> None:
        r"""Adds a history to the engine state.

        Args:
            history (``BaseHistory``): Specifies the history to
                add to the engine state.
            key (str or ``None``, optional): Specifies the key to
                store the history. If ``None``, the name of the
                history is used. Default: ``None``

        Example usage:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> from gravitorch.utils.history import MinScalarHistory
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.add_history(MinScalarHistory("loss"))
            >>> engine.add_history(MinScalarHistory("loss"), 'my key')
        """

    @abstractmethod
    def add_module(self, name: str, module: Any) -> None:
        r"""Adds a module to the engine.

        The state dict of the module will be added to the engine
        state when a checkpoint is created.

        Args:
            name (str): Specifies the name of the module.
            module: Specifies the module to add to the checkpoint. The
                module needs to have the ``state_dict`` and
                ``load_state_dict`` methods if you want to store the
                state of the module in the engine checkpoint. If the
                module does not have ``state_dict`` and
                ``load_state_dict`` methods, it will be ignored when
                the state dict is created or loaded.

        Example usage:

        .. code-block:: python

            >>> from torch import nn
            >>> from gravitorch.engines import BaseEngine
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.add_module('my_module', nn.Linear(4, 5))
        """

    @abstractmethod
    def create_artifact(self, artifact: BaseArtifact) -> None:
        r"""Creates an artifact.

        Args:
            artifact (``BaseArtifact``): Specifies the artifact to
            create.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> from gravitorch.utils.artifacts import JSONArtifact
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.create_artifact(JSONArtifact(tag='metric', data={'f1_score': 42}))
        """

    @abstractmethod
    def eval(self) -> None:
        r"""Evaluates the model on the given evaluation dataset with the given
        metrics/loss.

        The evaluation dataset has to be defined.

        Raises:
            RuntimeError: if the evaluation dataset is not be defined.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.eval()
        """

    @abstractmethod
    def fire_event(self, event: str) -> None:
        r"""Fires the handler(s) for the given event.

        Args:
            event (str): Specifies the event to fire.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> from gravitorch.utils.events import VanillaEventHandler
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.fire_event('my_event')  # should do nothing because there is no event handler
            >>> def hello_handler():
            ...     print("Hello!")
            >>> engine.add_event_handler('my_event', VanillaEventHandler(hello_handler))
            >>> engine.fire_event('my_event')
            Hello!
        """

    @abstractmethod
    def get_history(self, key: str) -> BaseHistory:
        r"""Gets the history associated to a key.

        Args:
            key (str): Specifies the key of the history to retrieve.

        Returns
            ``BaseHistory``: The history if it exists,
                otherwise it returns an empty history. The created
                empty history is a ``GenericHistory``.

        Example:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> from gravitorch.utils.history import MinScalarHistory
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.add_history(MinScalarHistory("loss"))
            >>> engine.get_history('loss')
            MinScalarHistory(name='loss', ...)
            >>> engine.get_history('new_history')
            GenericHistory(name='new_history', ...)
        """

    @abstractmethod
    def get_histories(self) -> dict[str, BaseHistory]:
        r"""Gets all histories in the state.

        Returns
            ``dict``: The histories with their keys.

        Example:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> from gravitorch.utils.history import MinScalarHistory
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.add_history(MinScalarHistory("loss"))
            >>> engine.get_histories()
            {'loss': MinScalarHistory(name='loss', ...)}
        """

    @abstractmethod
    def get_module(self, name: str) -> Any:
        r"""Gets a module.

        Args:
            name (str): Specifies the module to get.

        Returns:
            The module

        Raises:
            ValueError if the module does not exist.

        Example:

        .. code-block:: python

            >>> from torch import nn
            >>> from gravitorch.engines import BaseEngine
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.add_module('model', nn.Linear(4, 6))
            >>> engine.get_module('model')
            Linear(in_features=4, out_features=6, bias=True)
            >>> engine.get_module('missing_module')
            ValueError
        """

    @abstractmethod
    def has_event_handler(self, event_handler: BaseEventHandler, event: str | None = None) -> bool:
        r"""Indicates if a handler is registered in the event manager.

        Note that this method relies on the ``__eq__`` method of the
        given event handler to compare event handlers.

        Args:
            event_handler (``BaseEventHandler``): Specifies the event
                handler to check.
            event (str or ``None``): Specifies an event to check. If
                the value is ``None``, it will check all the events.
                Default: ``None``

        Example usage:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> from gravitorch.utils.events import VanillaEventHandler
            >>> engine: BaseEngine = ...  # Create an engine
            # Define a handler
            >>> def hello_handler():
            ...     print("Hello!")
            >>> engine.has_event_handler(VanillaEventHandler(hello_handler))
            False
            >>> engine.has_event_handler(VanillaEventHandler(hello_handler), 'my_event')
            False
            >>> engine.add_event_handler('my_event', VanillaEventHandler(hello_handler))
            >>> engine.has_event_handler(VanillaEventHandler(hello_handler))
            True
            >>> engine.has_event_handler(VanillaEventHandler(hello_handler), 'my_event')
            True
            >>> engine.has_event_handler(VanillaEventHandler(hello_handler), 'other_event')
            False
        """

    @abstractmethod
    def has_history(self, key: str) -> bool:
        r"""Indicates if the engine has a history for the given key.

        Args:
            key (str): Specifies the key of the history.

        Returns
            bool: ``True`` if the history exists, ``False`` otherwise.

        Example:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> from gravitorch.utils.history import MinScalarHistory
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.add_history(MinScalarHistory("loss"))
            >>> engine.has_history('loss')
            True
            >>> engine.has_history('missing_history')
            False
        """

    @abstractmethod
    def has_module(self, name: str) -> bool:
        r"""Indicates if the engine has a module for the given name.

        Args:
            name (str): Specifies the name of the module.

        Returns:
            bool: ``True`` if the module exists, otherwise ``False``.

        Example:

        .. code-block:: python

            >>> from torch import nn
            >>> from gravitorch.engines import BaseEngine
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.add_module('model', nn.Linear(4, 6))
            >>> engine.has_module('model')
            True
            >>> engine.has_module('missing_module')
            False
        """

    @abstractmethod
    def increment_epoch(self, increment: int = 1) -> None:
        r"""Increments the epoch value by the given value.

        Args:
            increment (int, optional): Specifies the increment for the
                epoch value. Default: ``1``

        Example usage:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.epoch
            1
            # Increment the epoch number by 1.
            >>> engine.increment_epoch()
            >>> engine.epoch
            0
            # Increment the epoch number by 10.
            >>> engine.increment_epoch(10)
            >>> engine.epoch
            10
        """

    @abstractmethod
    def increment_iteration(self, increment: int = 1) -> None:
        r"""Increments the iteration value by the given value.

        Args:
            increment (int, optional): Specifies the increment for
                the iteration value. Default: ``1``

        Example usage:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.iteration
            -1
            # Increment the iteration number by 1.
            >>> engine.increment_iteration()
            >>> engine.iteration
            0
            # Increment the iteration number by 10.
            >>> engine.increment_iteration(10)
            >>> engine.iteration
            10
        """

    @abstractmethod
    def load_state_dict(self, state_dict: dict) -> None:
        r"""Loads the state values from a dict.

        Args:
            state_dict (dict): a dict with parameters

        Example usage:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> engine: BaseEngine = ...  # Create an engine
            >>> state = {...}  # The content of the state dict depends on the engine.
            >>> engine.load_state_dict(state)
        """

    @abstractmethod
    def log_figure(
        self,
        key: str,
        figure: matplotlib.pyplot.Figure,  # noqa: F821
        step: Step | None = None,
    ) -> None:
        r"""Logs a figure for the given key and step.

        Args:
            key (str): Specifies the key used to identify the figure.
            figure (``matplotlib.pyplot.Figure``): Specifies the
                figure to log.
            step (``Step``, optional): Specifies the step value to
                record. If ``None``, it will use the default step.
                Default: ``None``

        Example usage:

        .. code-block:: python

            >>> import matplotlib.pyplot as plt
            >>> from gravitorch.engines import BaseEngine
            >>> from gravitorch.utils.exp_trackers import EpochStep, IterationStep
            >>> engine: BaseEngine = ...  # Create an engine
            >>> fig, axes = plt.subplots()
            >>> engine.log_figure('my_figure', fig)
            >>> engine.log_figure('my_figure', fig, EpochStep(2))
            >>> engine.log_figure('my_figure', fig, IterationStep(10))
        """

    @abstractmethod
    def log_figures(
        self,
        figures: dict[str, matplotlib.pyplot.Figure],  # noqa: F821
        step: Step | None = None,
    ) -> None:
        r"""Logs a dictionary of figures for a given step.

        Args:
            figures (dict): Specifies the dictionary of figures to log.
            step (``Step``, optional): Specifies the step value to
                record. If ``None``, it will use the default step.
                Default: ``None``

        Example usage:

        .. code-block:: python

            >>> import matplotlib.pyplot as plt
            >>> from gravitorch.engines import BaseEngine
            >>> from gravitorch.utils.exp_trackers import EpochStep, IterationStep
            >>> engine: BaseEngine = ...  # Create an engine
            >>> fig, axes = plt.subplots()
            >>> engine.log_figures({'my_figure_1': fig, 'my_figure_2': fig})
            >>> engine.log_figures({'my_figure_1': fig, 'my_figure_2': fig}, EpochStep(2))
            >>> engine.log_figures({'my_figure_1': fig, 'my_figure_2': fig}, IterationStep(10))
        """

    @abstractmethod
    def log_metric(self, key: str, value: int | float, step: Step | None = None):
        r"""Logs a single metric.

        It is possible to have access to the last logged values by
        using the ``get_history(key)`` method.

        Args:
            key (str): Specifies the key used to identify the scalar
                value.
            value (int or float): Specifies the value to log.
            step (``Step``, optional): Specifies the step value to
                record. If ``None``, it will use the default step.
                Default: ``None``

        Example usage:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> from gravitorch.utils.exp_trackers import EpochStep, IterationStep
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.log_metric(key='metric name', value=1.2)
            >>> engine.log_metric(key='metric name', value=1.2, step=EpochStep(2))
            >>> engine.log_metric(key='metric name', value=1.2, step=IterationStep(10))
        """

    @abstractmethod
    def log_metrics(self, metrics: dict[str, int | float], step: Step | None = None) -> None:
        r"""Logs a dictionary of multiple metrics.

        It is possible to have access to the last logged values by
        using the ``get_history(key)`` method.

        Args:
            metrics (dict): Specifies the dictionary of metric to log.
            step (``Step``, optional): Specifies the step value to
                record. If ``None``, it will use the default step.
                Default: ``None``

        Example usage:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> from gravitorch.utils.exp_trackers import EpochStep, IterationStep
            >>> engine: BaseEngine = ...  # Create an engine
            >>> my_metrics = {'metric1': 12, 'metric2': 0.35}
            >>> engine.log_metrics(metrics)
            >>> engine.log_metrics(my_metrics, EpochStep(2))
            >>> engine.log_metrics(my_metrics, IterationStep(10))
        """

    @abstractmethod
    def remove_event_handler(self, event: str, event_handler: BaseEventHandler) -> None:
        r"""Removes an event handler of a given event.

        Note that if the same event handler was added multiple times
        the event, all the duplicated handlers are removed. This
        method relies on the ``__eq__`` method of the input event
        handler to compare event handlers.

        Args:
            event (str): Specifies the event handler is attached to.
            event_handler (``BaseEventHandler``): Specifies the event
                handler to remove.

        Raises:
            ValueError: if the event does not exist or if the handler
                is not attached to the event.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> from gravitorch.utils.events import VanillaEventHandler
            >>> engine: BaseEngine = ...  # Create an engine
            >>> def hello_handler():
            ...     print('Hello!')
            >>> engine.add_event_handler('my_event', VanillaEventHandler(hello_handler))
            >>> engine.has_event_handler(VanillaEventHandler(hello_handler), 'my_event')
            True
            >>> engine.remove_event_handler('my_event', VanillaEventHandler(hello_handler))
            >>> engine.has_event_handler(VanillaEventHandler(hello_handler), 'my_event')
            False
        """

    @abstractmethod
    def remove_module(self, name: str) -> None:
        r"""Removes a module from the engine state.

        Args:
            name (str): Specifies the name of the module to remove.

        Raises:
            ValueError if the module name is not found.

        Example:

        .. code-block:: python

            >>> from torch import nn
            >>> from gravitorch.engines import BaseEngine
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.add_module('model', nn.Linear(4, 6))
            >>> engine.has_module('model')
            True
            >>> engine.remove_module('model')
            >>> engine.has_module('model')
            False
            >>> engine.remove_module('missing_model')
            ValueError
        """

    @abstractmethod
    def state_dict(self) -> dict[str, Any]:
        r"""Gets a dictionary containing a state of the engine.

        Note this state does not capture the handler states.

        Returns:
            dict: The dictionary containing a state of the engine.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> engine: BaseEngine = ...  # Create an engine
            >>> state = engine.state_dict()
        """

    @abstractmethod
    def terminate(self) -> None:
        r"""Sends terminate signal to the engine to stop the training after the
        current epoch.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.should_terminate
            False
            >>> engine.terminate()
            >>> engine.should_terminate
            True
        """

    @abstractmethod
    def train(self) -> None:
        r"""Trains the model on the given training dataset.

        If an evaluation dataset is given, the model will be evaluated
        on the evaluation dataset at each epoch.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.engines import BaseEngine
            >>> engine: BaseEngine = ...  # Create an engine
            >>> engine.train()
        """



__all__ = [
    "freeze_module",
    "get_module_device",
    "get_module_devices",
    "get_module_input_size",
    "get_module_name",
    "get_module_output_size",
    "has_batch_norm",
    "has_learnable_parameters",
    "has_parameters",
    "is_batch_first",
    "is_module_on_device",
    "module_mode",
    "num_learnable_parameters",
    "num_parameters",
    "top_module_mode",
    "unfreeze_module",
]

from collections.abc import Generator
from contextlib import contextmanager

import torch
from torch import nn


def has_parameters(module: nn.Module) -> bool:
    r"""Indicates if the module has parameters.

    Args:
        module (``torch.nn.Module``): Specifies the module to test.

    Returns:
        bool: ``True`` if the module has at least one parameter,
            ``False`` otherwise.
    """
    try:
        next(module.parameters())
        return True
    except StopIteration:
        return False


def has_learnable_parameters(module: nn.Module) -> bool:
    r"""Indicates if the module has learnable parameters.

    Args:
        module (``torch.nn.Module``): Specifies the module to test.

    Returns:
        bool: ``True`` if the module has at least one learnable
            parameter, ``False`` otherwise.
    """
    return num_learnable_parameters(module) > 0


def num_parameters(module: nn.Module) -> int:
    r"""Computes the number of parameters.

    Args:
        module (``torch.nn.Module``): Specifies the module to compute
            the number of parameters.

    Returns:
        int: The number of parameters.
    """
    return sum(params.numel() for params in module.parameters())


def num_learnable_parameters(module: nn.Module) -> int:
    r"""Computes the number of learnable parameters.

    Args:
        module (``torch.nn.Module``): Specifies the module to compute
            the number of learnable parameters..

    Returns:
        int: The number of learnable parameters.
    """
    return sum(params.numel() for params in module.parameters() if params.requires_grad)


def freeze_module(module: nn.Module) -> None:
    r"""Freezes the parameters of the given module.

    Args:
        module (``torch.nn.Module``): Specifies the module to freeze.
    """
    for param in module.parameters():
        param.requires_grad = False


def unfreeze_module(module: nn.Module) -> None:
    r"""Unfreezes the parameters of the given module.

    Args:
        module (``torch.nn.Module``): Specifies the module to
            unfreeze.
    """
    for param in module.parameters():
        param.requires_grad = True


def get_module_device(module: nn.Module) -> torch.device:
    r"""Get the device used by this module.

    This function assumes the module uses a single device. If the
    module uses several devices, you should use
    ``get_module_devices``. It returns ``torch.device('cpu')`` if
    the model does not have parameters.

    Args:
        module (``torch.nn.Module``): Specifies the module.

    Returns:
        ``torch.device``: The device
    """
    if not has_parameters(module):
        return torch.device("cpu")
    return next(module.parameters()).device


def get_module_devices(module: nn.Module) -> tuple[torch.device, ...]:
    r"""Get the devices used in a module.

    Args:
        module (``torch.nn.Module``): Specifies the module.

    Returns:
        tuple: The tuple of ``torch.device``s used in the module.
    """
    devices = set()
    for param in module.parameters():
        devices.add(param.device)
    return tuple(devices)


def is_module_on_device(module: nn.Module, device: torch.device) -> bool:
    r"""Indicates if all the parameters of a module are on the specified device.

    Args:
        module (``torch.nn.Module``): Specifies the module.
        device (``torch.device``): Specifies the device.

    Returns:
        bool: ``True`` if all the parameters of the module are on the
            specified device, otherwise ``False``.
    """
    return all([p.device == device for p in module.parameters()])


def get_module_input_size(module: nn.Module) -> int:
    r"""Gets the input size of a module.

    This function works only for the module with a single input size.

    Note: it is an experimental function that supports a limited
    number of modules:

        - ``ModuleList``: assumes all the modules have the same
            input size.

    Args:
        module (``torch.nn.Module``): Specifies the input module.

    Returns:
        int: The input size of the module.

    Raises:
        TypeError: if the module is not supported.
    """
    if hasattr(module, "input_size"):
        return module.input_size
    if hasattr(module, "in_features"):
        return module.in_features
    if hasattr(module, "in_channels"):
        return module.in_channels
    if isinstance(module, nn.Sequential):
        return _get_sequential_input_size(module)
    if isinstance(module, nn.ModuleList) and len(module) > 0:
        return get_module_input_size(module[0])
    if isinstance(module, nn.MultiheadAttention):
        return module.embed_dim
    if isinstance(module, (nn.TransformerEncoderLayer, nn.TransformerDecoderLayer)):
        return get_module_input_size(module.self_attn)
    if isinstance(module, (nn.TransformerEncoder, nn.TransformerDecoder)):
        return get_module_input_size(module.layers)

    raise TypeError(f"{type(module)} module is not supported")


def _get_sequential_input_size(sequential: nn.Sequential) -> int:
    r"""Gets the input size of a ``torch.nn.Sequential``.

    This function finds the first layer of the ``torch.nn.Sequential``
    for which it is possible to compute the input size.

    Args:
        sequential (``torch.nn.Sequential``): Specifies the input
            module.

    Returns:
        int: The input size of the ``torch.nn.Sequential`` object
            if possible.

    Raises:
        TypeError if the input is not a ``torch.nn.Sequential`` or
            if none of the child modules are supported.
    """
    if not isinstance(sequential, nn.Sequential):
        raise TypeError(
            f"Only the `torch.nn.Sequential` inputs are supported. (received: {sequential})"
        )
    for module in sequential:
        try:
            return get_module_input_size(module)
        except TypeError:
            pass
    raise TypeError(
        "It is not possible to find the input size because none of the child modules are supported"
    )


def get_module_output_size(module: nn.Module) -> int:
    r"""Gets the output size of a module.

    This function works only for the module with a single output size.

    Note: it is an experimental function that supports a limited
    number of modules:

        - ``ModuleList``: assumes all the modules have the same
            output  size.

    Args:
        module (``torch.nn.Module``): Specifies the input module.

    Returns:
        int: The output size of the module if possible.

    Raises:
        TypeError: if the module is not supported.
    """
    if hasattr(module, "output_size"):
        return module.output_size
    if hasattr(module, "out_features"):
        return module.out_features
    if hasattr(module, "out_channels"):
        return module.out_channels
    if isinstance(module, (nn.RNN, nn.LSTM, nn.GRU)):
        return module.hidden_size
    if isinstance(module, nn.Embedding):
        return module.embedding_dim
    if isinstance(module, nn.Sequential):
        return _get_sequential_output_size(module)
    if isinstance(module, nn.ModuleList) and len(module) > 0:
        return get_module_output_size(module[0])
    if isinstance(module, nn.MultiheadAttention):
        return module.embed_dim
    if isinstance(module, (nn.TransformerEncoderLayer, nn.TransformerDecoderLayer)):
        return get_module_output_size(module.self_attn)
    if isinstance(module, (nn.TransformerEncoder, nn.TransformerDecoder)):
        return get_module_output_size(module.layers)

    raise TypeError(f"{type(module)} module is not supported")


def _get_sequential_output_size(sequential: nn.Sequential) -> int:
    r"""Gets the output size of a ``torch.nn.Sequential``.

    This function finds the last layer of the ``torch.nn.Sequential``
    for which it is possible to compute the output size.

    Args:
        sequential (``torch.nn.Sequential``): Specifies the input
            module.

    Returns:
        int: The output size of the ``torch.nn.Sequential`` object
            if possible.

    Raises:
        TypeError if the input is not a ``torch.nn.Sequential`` or
            if none of the child modules are supported.
    """
    if not isinstance(sequential, nn.Sequential):
        raise TypeError(
            f"Only the `torch.nn.Sequential` inputs are supported. (received: {sequential})"
        )
    for module in sequential[::-1]:
        try:
            return get_module_output_size(module)
        except TypeError:
            pass
    raise TypeError(
        "It is not possible to find the output size because none of the child modules are supported"
    )


def has_batch_norm(module: nn.Module) -> bool:
    r"""Indicates if the module has at least one batch norm layer.

    This function only verifies if the module has at least one of the
    following layer:

        - ``torch.nn.BatchNorm1d``
        - ``torch.nn.BatchNorm2d``
        - ``torch.nn.BatchNorm3d``
        - ``torch.nn.SyncBatchNorm``

    Args:
        module (``torch.nn.Module``): Specifies the input module.

    Returns:
        bool: ``True`` if the module has at least one batch norm layer,
            otherwise ``False``
    """
    for layer in module.modules():
        if isinstance(layer, (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.SyncBatchNorm)):
            return True
    return False


def get_module_name(layer: nn.Module) -> str:
    r"""Gets the name of a module.

    The name of the module is the class name lower case.

    Args:
        layer (``torch.nn.Module``): Specifies the layer.

    Returns:
        str: The layer name.
    """
    return layer.__class__.__qualname__


def is_batch_first(module: nn.Module) -> bool:
    r"""Indicates if the input and output tensors are provided as ``(batch_size,
    seq_len, *)`` instead of ``(seq_len, batch_size,

    *)``.

    Args:
        module (``torch.nn.Module``): Specifies the module to check.

    Returns:
        bool: ``True`` if batch first, otherwise ``False``.

    Example usage:

    .. code-block:: python

        >>> from torch import nn
        >>> from gravitorch.nn import is_batch_first
        >>> is_batch_first(nn.MultiheadAttention(4, 1, batch_first=False))
        False
        >>> is_batch_first(nn.MultiheadAttention(4, 1, batch_first=True))
        True
    """
    if hasattr(module, "batch_first"):
        return module.batch_first
    if isinstance(module, (nn.TransformerEncoderLayer, nn.TransformerDecoderLayer)):
        return is_batch_first(module.self_attn)
    if isinstance(module, (nn.TransformerEncoder, nn.TransformerDecoder)):
        return is_batch_first(module.layers[0])
    raise TypeError(f"{type(module)} module is not supported")


@contextmanager
def module_mode(module: nn.Module) -> Generator[None, None, None]:
    r"""Implements a context manager that restores the mode (train or eval) of
    every submodule individually.

    This context manager only restores the mode at the top-level.

    Args:
        module (``torch.nn.Module``): Specifies the module to restore
            the mode.

    Example usage:

    .. code-block:: python

        >>> from torch import nn
        >>> from gravitorch.nn import module_mode
        >>> module = nn.ModuleDict({'module1': nn.Linear(4, 6), 'module2': nn.Linear(2, 4).eval()})
        >>> print(module['module1'].training, module['module2'].training)
        True False
        >>> with module_mode(module):
        ...     module.eval()
        ...     print(module['module1'].training, module['module2'].training)
        False False
        >>> print(module['module1'].training, module['module2'].training)
        True False
    """
    modes = {}
    for name, submodule in module.named_modules():
        modes[name] = submodule.training
    try:
        yield
    finally:
        for name, submodule in module.named_modules():
            submodule.train(modes[name])


@contextmanager
def top_module_mode(module: nn.Module) -> Generator[None, None, None]:
    r"""Implements a context manager that restores the mode (train or eval) of a
    given module.

    This context manager only restores the mode at the top-level.

    Args:
        module (``torch.nn.Module``): Specifies the module to restore
            the mode.

    Example usage:

    .. code-block:: python

        >>> from torch import nn
        >>> from gravitorch.nn import module_mode
        >>> module = nn.Linear(4, 6)
        >>> print(module.training)
        True
        >>> with top_module_mode(module):
        ...     module.eval()
        ...     print(module)
        False
        >>> print(module.training)
        True
    """
    mode = module.training
    try:
        yield
    finally:
        module.train(mode)



__all__ = ["DataDistributedParallelModelCreator", "to_ddp"]

import logging
from typing import Optional, Union

from torch import nn
from torch.nn.parallel import DistributedDataParallel

from gravitorch import distributed as dist
from gravitorch.creators.model.base import BaseModelCreator, setup_model_creator
from gravitorch.engines.base import BaseEngine
from gravitorch.utils.format import str_add_indent, to_pretty_json_str

logger = logging.getLogger(__name__)


class DataDistributedParallelModelCreator(BaseModelCreator):
    r"""Implements a model creator that wraps a created model with
    ``DistributedDataParallel``.

    Args:
        model_creator (``BaseModelCreator`` or dict): Specifies a
            model creator or its configuration. The created model
            should be compatible with ``DistributedDataParallel``.
        ddp_kwargs (dict or ``None``): Specifies some keyword
            arguments used to instantiate the
            ``DistributedDataParallel``. Please read the documentation
            of ``DistributedDataParallel`` to see the possible
            options. Note that it is not possible to set ``module``
            and ``device_ids`` with a keyword argument.
            Default: ``None``
    """

    def __init__(
        self, model_creator: Union[BaseModelCreator, dict], ddp_kwargs: Optional[dict] = None
    ):
        self._model_creator = setup_model_creator(model_creator)
        self._ddp_kwargs = ddp_kwargs or {}

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  model_creator={self._model_creator},\n"
            f"  ddp_kwargs={str_add_indent(to_pretty_json_str(self._ddp_kwargs))},\n"
            ")"
        )

    def create(self, engine: BaseEngine) -> nn.Module:
        model = self._model_creator.create(engine)
        return to_ddp(module=model, ddp_kwargs=self._ddp_kwargs)


def to_ddp(module: nn.Module, ddp_kwargs: Optional[dict] = None) -> nn.Module:
    r"""Wraps a module with the ``DistributedDataParallel`` module.

    Args:
        module (``torch.nn.Module``): Specifies the module to wrap
            with ``DistributedDataParallel``. The module should be
            compatible with ``DistributedDataParallel``. If you use
            NCCL, the module should be on a CUDA device.
        ddp_kwargs (dict or ``None``): Specifies some keyword
            arguments used to instantiate the
            ``DistributedDataParallel``. Please read the
            documentation of ``DistributedDataParallel`` to see the
            possible options. Note that it is not possible to set
            ``module`` and ``device_ids`` with a keyword argument.
            Default: ``None``

    Returns:
        ``torch.nn.Module``: The model wrapped in a
            ``DistributedDataParallel`` module.
    """
    if isinstance(module, DistributedDataParallel):
        logger.warning(
            "No operation is performed because the module is already a DistributedDataParallel"
        )
        return module
    ddp_kwargs = ddp_kwargs or {}
    backend = dist.backend()
    if backend == dist.Backend.NCCL:
        lrank = dist.get_local_rank()
        logger.info(f"Applying DistributedDataParallel on module, device id: {lrank}")
        return DistributedDataParallel(module, device_ids=[lrank], **ddp_kwargs)
    if backend == dist.Backend.GLOO:
        logger.info("Applying DistributedDataParallel on module")
        return DistributedDataParallel(module, **ddp_kwargs)
    return module



__all__ = [
    "Asinh",
    "AsinhBarronRobustLoss",
    "AsinhCosSinScalarEncoder",
    "AsinhMSELoss",
    "AverageFusion",
    "BarronRobustLoss",
    "BinaryFocalLoss",
    "BinaryFocalLossWithLogits",
    "Clamp",
    "ClampLoss",
    "ConcatFusion",
    "CosSinScalarEncoder",
    "ExpSin",
    "ExU",
    "FlattenModule",
    "FusionFFN",
    "FusionNorm",
    "Gaussian",
    "Isymlog",
    "Laplacian",
    "Log1p",
    "AsinhScalarEncoder",
    "AsinhCosSinScalarEncoder",
    "MSLELoss",
    "ModuleSummary",
    "MultiQuadratic",
    "MulticlassFlattenModule",
    "MultiplicationFusion",
    "OnePolynomial",
    "Quadratic",
    "ReLUn",
    "RelativeMSELoss",
    "RelativeSmoothL1Loss",
    "ResidualBlock",
    "Safeexp",
    "Safelog",
    "ScalarEncoderFFN",
    "ScaleShiftSequenceGaussianRFF",
    "SequenceGaussianRFF",
    "SequenceShiftScale",
    "SequenceToBatch",
    "ShiftScale",
    "Sin",
    "Sinh",
    "Squeeze",
    "SumFusion",
    "Symlog",
    "SymlogMSELoss",
    "SymmetricRelativeSmoothL1Loss",
    "ToBinaryLabel",
    "ToCategoricalLabel",
    "ToFloat",
    "ToLong",
    "WarmupSequenceLoss",
    "compute_parameter_stats",
    "create_sequential",
    "find_module_state_dict",
    "freeze_module",
    "get_module_device",
    "get_module_devices",
    "get_module_input_size",
    "get_module_name",
    "get_module_output_size",
    "has_batch_norm",
    "has_learnable_parameters",
    "has_parameters",
    "is_batch_first",
    "is_module_on_device",
    "load_checkpoint_to_module",
    "module_mode",
    "num_learnable_parameters",
    "num_parameters",
    "setup_nn_module",
    "show_parameter_stats",
    "state_dicts_are_equal",
    "top_module_mode",
    "unfreeze_module",
]

from gravitorch.nn.activations import ReLUn
from gravitorch.nn.alpha_activations import (
    ExpSin,
    Gaussian,
    Laplacian,
    MultiQuadratic,
    Quadratic,
    Sin,
)
from gravitorch.nn.barron_loss import AsinhBarronRobustLoss, BarronRobustLoss
from gravitorch.nn.clamp import Clamp, ClampLoss
from gravitorch.nn.creators import create_sequential
from gravitorch.nn.exu import ExU
from gravitorch.nn.flatten import FlattenModule, MulticlassFlattenModule
from gravitorch.nn.focal_loss import BinaryFocalLoss, BinaryFocalLossWithLogits
from gravitorch.nn.fourier_feature import (
    ScaleShiftSequenceGaussianRFF,
    SequenceGaussianRFF,
)
from gravitorch.nn.fusion import (
    AverageFusion,
    ConcatFusion,
    FusionFFN,
    FusionNorm,
    MultiplicationFusion,
    SumFusion,
)
from gravitorch.nn.residual import ResidualBlock
from gravitorch.nn.robust_loss import (
    AsinhMSELoss,
    MSLELoss,
    RelativeMSELoss,
    RelativeSmoothL1Loss,
    SymlogMSELoss,
    SymmetricRelativeSmoothL1Loss,
)
from gravitorch.nn.scalar_encoding import (
    AsinhCosSinScalarEncoder,
    AsinhScalarEncoder,
    CosSinScalarEncoder,
    ScalarEncoderFFN,
)
from gravitorch.nn.shift_scale import SequenceShiftScale, ShiftScale
from gravitorch.nn.transforms import (
    Asinh,
    Isymlog,
    Log1p,
    OnePolynomial,
    Safeexp,
    Safelog,
    SequenceToBatch,
    Sinh,
    Squeeze,
    Symlog,
    ToBinaryLabel,
    ToCategoricalLabel,
    ToFloat,
    ToLong,
)
from gravitorch.nn.utils import (
    ModuleSummary,
    compute_parameter_stats,
    find_module_state_dict,
    freeze_module,
    get_module_device,
    get_module_devices,
    get_module_input_size,
    get_module_name,
    get_module_output_size,
    has_batch_norm,
    has_learnable_parameters,
    has_parameters,
    is_batch_first,
    is_module_on_device,
    load_checkpoint_to_module,
    module_mode,
    num_learnable_parameters,
    num_parameters,
    setup_nn_module,
    show_parameter_stats,
    state_dicts_are_equal,
    top_module_mode,
    unfreeze_module,
)
from gravitorch.nn.warmup_seq_loss import WarmupSequenceLoss



__all__ = ["CategoricalCrossEntropy"]

import logging
from typing import Union

from torch import Tensor
from torch.nn.functional import cross_entropy

from gravitorch.models.metrics.base_epoch import BaseStateEpochMetric
from gravitorch.models.metrics.state import BaseState, MeanErrorState

logger = logging.getLogger(__name__)


class CategoricalCrossEntropy(BaseStateEpochMetric):
    r"""Implements a metric to compute the categorical cross-entropy.

    Args:
        mode (str): Specifies the mode.
        name (str, optional): Specifies the name used to log the
            metric. Default: ``'cat_acc'``
        state (``BaseState`` or dict or ``None``, optional): Specifies
            the metric state or its configuration. If ``None``,
            ``MeanErrorState`` is instantiated. Default: ``None``
    """

    def __init__(
        self,
        mode: str,
        name: str = "cat_ce",
        state: Union[BaseState, dict, None] = None,
    ):
        super().__init__(mode=mode, name=name, state=state or MeanErrorState())

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the metric given a mini-batch of examples.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn, num_classes)`` and type float):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` or ``(d0, d1, ..., dn, 1)`` and
                type long or float): Specifies the categorical
                targets. The values have to be in
                ``{0, 1, ..., num_classes-1}``.
        """
        self._state.update(
            cross_entropy(
                prediction.flatten(start_dim=0, end_dim=-2),
                target.flatten().long(),
                reduction="none",
            )
        )



r"""This package contains some utility functions for the models."""

__all__ = [
    "ModelSummary",
    "analyze_model_architecture",
    "analyze_model_network_architecture",
    "analyze_module_architecture",
    "attach_module_to_engine",
    "is_loss_decreasing",
    "is_loss_decreasing_with_adam",
    "is_loss_decreasing_with_sgd",
    "setup_and_attach_model",
    "setup_model",
]

from gravitorch.models.utils.architecture_analysis import (
    analyze_model_architecture,
    analyze_model_network_architecture,
    analyze_module_architecture,
)
from gravitorch.models.utils.setup_and_attach import (
    attach_module_to_engine,
    setup_and_attach_model,
    setup_model,
)
from gravitorch.models.utils.summary import ModelSummary
from gravitorch.models.utils.testing import (
    is_loss_decreasing,
    is_loss_decreasing_with_adam,
    is_loss_decreasing_with_sgd,
)



__all__ = ["setup_optimizer_creator"]

import logging
from typing import Union

from gravitorch.creators.optimizer.base import BaseOptimizerCreator
from gravitorch.creators.optimizer.noo import NoOptimizerCreator
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def setup_optimizer_creator(
    creator: Union[BaseOptimizerCreator, dict, None]
) -> BaseOptimizerCreator:
    r"""Sets up the optimizer creator.

    The optimizer creator is instantiated from its configuration
    by using the ``BaseOptimizerCreator`` factory function.

    Args:
        creator (``BaseOptimizerCreator`` or dict or ``None``):
            Specifies the optimizer creator or its configuration.
            If ``None``, a ``NoOptimizerCreator`` is created.

    Returns:
        ``BaseOptimizerCreator``: The instantiated optimizer creator.
    """
    if creator is None:
        creator = NoOptimizerCreator()
    if isinstance(creator, dict):
        logger.info(
            "Initializing the optimizer creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseOptimizerCreator.factory(**creator)
    return creator



r"""IO-related utility functions."""

__all__ = [
    "load_json",
    "load_pickle",
    "load_text",
    "load_yaml",
    "save_json",
    "save_pickle",
    "save_pytorch",
    "save_text",
    "save_yaml",
]

import json
import logging
import pickle
from pathlib import Path
from typing import Any, Union

import numpy as np
import torch
import yaml
from torch import Tensor

logger = logging.getLogger(__name__)


def numpy_array_yaml_representer(dumper: yaml.Dumper, data: Union[np.ndarray, Tensor]):
    r"""This function computes the yaml representation of a numpy array or
    torch tensor."""
    if data.ndim == 0:
        scalar = data.item()
        if isinstance(scalar, float):
            return dumper.represent_float(scalar)
        return dumper.represent_int(scalar)
    return dumper.represent_list(data.tolist())


def numpy_datetime_yaml_representer(dumper: yaml.Dumper, data: np.datetime64):
    r"""This function computes the yaml representation of a numpy
    datetime64."""
    return dumper.represent_str(str(data))


def numpy_float_yaml_representer(dumper: yaml.Dumper, data: np.floating):
    r"""This function computes the yaml representation of a numpy
    float64/float32."""
    return dumper.represent_float(float(data))


def numpy_int_yaml_representer(dumper: yaml.Dumper, data: np.integer):
    r"""This function computes the yaml representation of a numpy
    int64/int32."""
    return dumper.represent_int(int(data))


# Set up the yaml dumper to deal with some data types like PyTorch tensors.
yaml.Dumper.add_representer(np.ndarray, numpy_array_yaml_representer)
yaml.Dumper.add_representer(np.datetime64, numpy_datetime_yaml_representer)
yaml.Dumper.add_representer(np.float64, numpy_float_yaml_representer)
yaml.Dumper.add_representer(np.float32, numpy_float_yaml_representer)
yaml.Dumper.add_representer(np.int64, numpy_int_yaml_representer)
yaml.Dumper.add_representer(np.int32, numpy_int_yaml_representer)
yaml.Dumper.add_representer(Tensor, numpy_array_yaml_representer)


###############
#    Text     #
###############


def load_text(path: Path) -> str:
    r"""Reads the data from a given text file.

    Args:
        path (``pathlib.Path``): Specifies the path where to the text
            file.

    Returns:
        The data from the text file.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from gravitorch.utils.io import load_text
        >>> data = load_text(Path('/path/to/data.txt'))
    """
    logger.debug(f"read {path}")
    with open(path) as file:
        data = file.read()
    return data


def save_text(to_save: str, path: Path) -> None:
    r"""Saves the given data in a text file.

    Args:
        to_save: Specifies the data to write in a text file.
        path (``pathlib.Path``): Specifies the path where to write the
            text file.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from gravitorch.utils.io import save_text
        >>> save_text("abc", Path('/path/to/data.txt'))
    """
    logger.debug(f"write data in a text file: {path}")
    path.parents[0].mkdir(exist_ok=True, parents=True)
    # Save to tmp, then commit by moving the file in case the job gets
    # interrupted while writing the file
    tmp_path = path.parents[0].joinpath(f"{path.name}.tmp")
    with open(tmp_path, mode="w") as file:
        file.write(to_save)
    tmp_path.rename(path)


################
#     JSON     #
################


class TensorJSONEncoder(json.JSONEncoder):
    r"""Defines a custom json encoder to work with numpy arrays and tensors."""

    def default(self, o: Any) -> Any:
        if isinstance(o, (Tensor, np.ndarray)):
            return o.tolist()
        if isinstance(o, np.integer):
            return int(o)
        if isinstance(o, np.floating):
            return float(o)
        return str(o)  # use the string representation


def load_json(path: Path) -> dict:
    r"""Loads the data from a given json file.

    Args:
        path (``pathlib.Path``): Specifies the path to the json file.

    Returns:
        dict: The data from the json file.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from gravitorch.utils.io import load_json
        >>> data = load_json(Path('/path/to/data.json'))
    """
    logger.debug(f"Loading data from {path}...")
    with open(path, mode="rb") as file:
        data = json.load(file)
    return data


def save_json(to_save: Any, path: Path) -> None:
    r"""Saves the given data in a json file.

    Note: the keys are automatically converted to a string.

    Args:
        to_save: Specifies the data to write in a json file. The data
            should be JSON compatible.
        path (``pathlib.Path``): Specifies the path where to write the
            json file.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from gravitorch.utils.io import save_json
        >>> save_json({"key": "value"}, Path('/path/to/data.json'))
    """
    logger.debug(f"Saving data in a JSON file: {path}")
    path.parents[0].mkdir(exist_ok=True, parents=True)
    # Save to tmp, then commit by moving the file in case the job gets
    # interrupted while writing the file
    tmp_path = path.parents[0].joinpath(f"{path.name}.tmp")
    with open(tmp_path, "w") as file:
        json.dump(to_save, file, sort_keys=False, cls=TensorJSONEncoder)
    tmp_path.rename(path)


##################
#     Pickle     #
##################


def load_pickle(path: Path) -> Any:
    r"""Loads the data from a given pickle file.

    Args:
        path (``pathlib.Path``): Specifies the path to the pickle
            file.

    Returns:
        The data from the pickle file.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from gravitorch.utils.io import load_pickle
        >>> data = load_pickle(Path('/path/to/data.pkl'))
    """
    logger.debug(f"Loading data from {path}...")
    with open(path, mode="rb") as file:
        data = pickle.load(file)
    return data


def save_pickle(to_save: Any, path: Path, protocol: int = pickle.HIGHEST_PROTOCOL) -> None:
    r"""Saves the given data in a pickle file.

    Args:
        to_save: Specifies the data to write in a pickle file.
        path (``pathlib.Path``): Specifies the path where to write
            the pickle file.
        protocol (int, optional): Specifies the pickle protocol. By
            default, it uses the highest protocol available.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from gravitorch.utils.io import save_pickle
        >>> save_pickle({"key": "value"}, Path('/path/to/data.pkl'))
    """
    logger.debug(f"Saving data in pickle file: {path}  (protocol={protocol})")
    path.parents[0].mkdir(exist_ok=True, parents=True)
    # Save to tmp, then commit by moving the file in case the job gets
    # interrupted while writing the file
    tmp_path = path.parents[0].joinpath(f"{path.name}.tmp")
    with open(tmp_path, mode="wb") as file:
        pickle.dump(to_save, file, protocol=protocol)
    tmp_path.rename(path)


###################
#     PyTorch     #
###################


def save_pytorch(to_save: Any, path: Path) -> None:
    r"""Saves the data in a PyTorch file.

    Args:
        to_save: Specifies the data to write in the PyTorch file.
        path (``pathlib.Path``): Specifies the path where to write the
            PyTorch file.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from gravitorch.utils.io import save_pytorch
        >>> save_pytorch({"key": "value"}, Path('/path/to/data.pt'))
    """
    logger.debug(f"Saving data in a PyTorch file: {path}")
    path.parents[0].mkdir(exist_ok=True, parents=True)
    # Save to tmp, then commit by moving the file in case the job gets
    # interrupted while writing the file
    tmp_path = path.parents[0].joinpath(f"{path.name}.tmp")
    torch.save(to_save, tmp_path)
    tmp_path.rename(path)


################
#     YAML     #
################


def load_yaml(path: Path) -> Any:
    r"""Loads the data from a given yaml file.

    Args:
        path (``pathlib.Path``): Specifies the path to the yaml file.

    Returns:
        The data from the yaml file.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from gravitorch.utils.io import load_yaml
        >>> data = load_yaml(Path('/path/to/data.yaml'))
    """
    logger.debug(f"Loading data from {path}...")
    with open(path, mode="rb") as file:
        data = yaml.safe_load(file)
    return data


def save_yaml(to_save: Any, path: Path) -> None:
    r"""Saves the data in a yaml file.

    Args:
        to_save: Specifies the data to write in a yaml file.
        path (``pathlib.Path``): Specifies the path where to write the
            yaml file.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from gravitorch.utils.io import save_yaml
        >>> save_yaml({"key": "value"}, Path('/path/to/data.yaml'))
    """
    logger.debug(f"Saving data in a yaml file: {path}")
    path.parents[0].mkdir(exist_ok=True, parents=True)
    # Save to tmp, then commit by moving the file in case the job gets
    # interrupted while writing the file
    tmp_path = path.parents[0].joinpath(f"{path.name}.tmp")
    with open(tmp_path, mode="w") as file:
        yaml.dump(to_save, file, Dumper=yaml.Dumper)
    tmp_path.rename(path)



r"""This module implements some concatenation based fusion layers."""

__all__ = ["ConcatFusion"]

import torch
from torch import Tensor
from torch.nn import Module


class ConcatFusion(Module):
    r"""Implements a module to concatenate inputs.

    Args:
        dim (int, optional): Specifies the fusion dimension. ``-1``
            means the last dimension. Default: ``-1``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn import ConcatFusion
        >>> module = ConcatFusion()
        >>> module
        ConcatFusion(dim=-1)
        >>> x1 = torch.tensor([[2, 3, 4], [5, 6, 7]], dtype=torch.float, requires_grad=True)
        >>> x2 = torch.tensor([[12, 13, 14], [15, 16, 17]], dtype=torch.float, requires_grad=True)
        >>> out = module(x1, x2)
        tensor([[ 2.,  3.,  4., 12., 13., 14.],
                [ 5.,  6.,  7., 15., 16., 17.]], grad_fn=<CatBackward0>)
        >>> out.mean().backward()
    """

    def __init__(self, dim: int = -1):
        super().__init__()
        self._dim = dim

    def extra_repr(self) -> str:
        return f"dim={self._dim}"

    def forward(self, *inputs: Tensor) -> Tensor:
        r"""Concatenates the list or tuple of inputs and then applied a feed-
        forward network (FFN) on the fused representation.

        Args:
            *inputs (list or tuple of ``torch.Tensor``): Specifies the
                tensors to concatenate.

        Returns:
            ``torch.Tensor``: The fused representation.
        """
        return torch.cat(inputs, dim=self._dim)



r"""This module defines some absolute relative error metrics."""

__all__ = ["AbsoluteRelativeError", "SymmetricAbsoluteRelativeError"]

import logging
from typing import Union

from torch import Tensor

from gravitorch.models.metrics.base_epoch import BaseStateEpochMetric
from gravitorch.models.metrics.state import BaseState, ErrorState
from gravitorch.nn.functional import (
    absolute_relative_error,
    symmetric_absolute_relative_error,
)

logger = logging.getLogger(__name__)


class AbsoluteRelativeError(BaseStateEpochMetric):
    r"""Implements the absolute relative error metric.

    Args:
        mode (str): Specifies the mode (e.g ``'train'`` or ``'eval'``).
        name (str, optional): Specifies the name of the metric. The
            name is used to log the metric results.
            Default: ``'abs_rel_err'``
        eps (float, optional): Specifies an arbitrary small strictly
            positive number to avoid undefined results when the target
            is zero. Default: ``1e-8``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics import AbsoluteRelativeError
        >>> metric = AbsoluteRelativeError('eval')
        >>> metric
        AbsoluteRelativeErrorMetric(mode=eval, name=abs_rel_err, eps=1e-08)
        >>> metric(torch.ones(2, 4), torch.ones(2, 4))
        >>> metric.value()
        {'eval/abs_rel_err_mean': 0.0,
         'eval/abs_rel_err_min': 0.0,
         'eval/abs_rel_err_max': 0.0,
         'eval/abs_rel_err_sum': 0.0,
         'eval/abs_rel_err_num_predictions': 8}
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/abs_rel_err_mean': 0.16666666666666666,
         'eval/abs_rel_err_min': 0.0,
         'eval/abs_rel_err_max': 1.0,
         'eval/abs_rel_err_sum': 2.0,
         'eval/abs_rel_err_num_predictions': 12}
        >>> metric.reset()
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/abs_rel_err_mean': 0.5,
         'eval/abs_rel_err_min': 0.0,
         'eval/abs_rel_err_max': 1.0,
         'eval/abs_rel_err_sum': 2.0,
         'eval/abs_rel_err_num_predictions': 4}
    """

    def __init__(
        self,
        mode: str,
        name: str = "abs_rel_err",
        eps: float = 1e-8,
        state: Union[BaseState, dict, None] = None,
    ):
        super().__init__(mode=mode, name=name, state=state or ErrorState())
        if eps <= 0:
            raise ValueError(
                f"Incorrect eps ({eps}). eps has to be an arbitrary small strictly "
                f"positive number to avoid undefined results when the target is zero."
            )
        self._eps = float(eps)

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the mean absolute percentage error metric given a mini-batch
        of examples.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the predictions.
            target (``torch.Tensor`` with same shape and data type as
                ``prediction``): Specifies the target tensor.
        """
        self._state.update(absolute_relative_error(prediction, target, eps=self._eps))


class SymmetricAbsoluteRelativeError(BaseStateEpochMetric):
    r"""Implements the symmetric absolute relative error (SARE) metric.

    This metric tracks the mean, maximum and minimum absolute
    relative error values.

    Args:
        mode (str): Specifies the mode (e.g ``'train'`` or ``'eval'``).
        name (str, optional): Specifies the name of the metric. The
            name is used to log the metric results.
            Default: ``'sym_abs_rel_err'``
        eps (float, optional): Specifies an arbitrary small strictly
            positive number to avoid undefined results when the target
            is zero. Default: ``1e-8``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics import SymmetricAbsoluteRelativeError
        >>> metric = SymmetricAbsoluteRelativeError('eval')
        >>> metric
        SymmetricAbsoluteRelativeErrorMetric(mode=eval, name=sym_abs_rel_err, eps=1e-08)
        >>> metric(torch.ones(2, 4), torch.ones(2, 4))
        >>> metric.value()
        {'eval/sym_abs_rel_err_mean': 0.0,
         'eval/sym_abs_rel_err_min': 0.0,
         'eval/sym_abs_rel_err_max': 0.0,
         'eval/sym_abs_rel_err_sum': 0.0,
         'eval/sym_abs_rel_err_num_predictions': 8}
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/sym_abs_rel_err_mean': 0.3333333333333333,
         'eval/sym_abs_rel_err_min': 0.0,
         'eval/sym_abs_rel_err_max': 2.0,
         'eval/sym_abs_rel_err_sum': 4.0,
         'eval/sym_abs_rel_err_num_predictions': 12}
        >>> metric.reset()
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/sym_abs_rel_err_mean': 1.0,
         'eval/sym_abs_rel_err_min': 0.0,
         'eval/sym_abs_rel_err_max': 2.0,
         'eval/sym_abs_rel_err_sum': 4.0,
         'eval/sym_abs_rel_err_num_predictions': 4}
    """

    def __init__(
        self,
        mode: str,
        name: str = "sym_abs_rel_err",
        eps: float = 1e-8,
        state: Union[BaseState, dict, None] = None,
    ):
        super().__init__(mode=mode, name=name, state=state or ErrorState())
        if eps <= 0:
            raise ValueError(
                f"Incorrect eps ({eps}). eps has to be an arbitrary small strictly "
                f"positive number to avoid undefined results when the target is zero."
            )
        self._eps = float(eps)

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the mean absolute percentage error metric given a mini-batch
        of examples.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the target tensor.
        """
        self._state.update(symmetric_absolute_relative_error(prediction, target, eps=self._eps))



__all__ = ["LogCudaMemory", "LogSysInfo"]

import logging
from types import TracebackType
from typing import Optional

from gravitorch.rsrc.base import BaseResource
from gravitorch.utils.cudamem import log_max_cuda_memory_allocated
from gravitorch.utils.sysinfo import log_system_info

logger = logging.getLogger(__name__)


class LogCudaMemory(BaseResource):
    r"""Implements a context manager to log the CUDA memory."""

    def __enter__(self) -> "LogCudaMemory":
        log_max_cuda_memory_allocated()
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        log_max_cuda_memory_allocated()

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"


class LogSysInfo(BaseResource):
    r"""Implements a context manager to log system information."""

    def __enter__(self) -> "LogSysInfo":
        log_system_info()
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        log_system_info()

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"



r"""This module implements a data summary for ``torch.Tensor``s of type
float."""

__all__ = ["FloatTensorDataSummary", "FloatTensorSequenceDataSummary"]

from typing import Union

from torch import Tensor

from gravitorch.utils.data_summary.continuous import (
    DEFAULT_QUANTILES,
    BaseContinuousDataSummary,
)
from gravitorch.utils.data_summary.sequence import BaseSequenceDataSummary


class FloatTensorDataSummary(BaseContinuousDataSummary[Tensor]):
    r"""Implements a data summary for ``torch.Tensor``s of type float.

    This data summary assumes that the data are continuous numerical
    values. This data summary computes the following statistics:

        - ``count``: the number of values
        - ``sum``: the sum of tne values
        - ``mean``: the mean of tne values
        - ``median``: the median of tne values
        - ``std``: the standard deviation of tne values
        - ``max``: the max value
        - ``min``: the min value
        - ``quantiles``: the quantile values

    Args:
        max_size (int, optional): Specifies the maximum size used to
            store the last values because it may not be possible to
            store all the values. This parameter is used to compute
            the median and the quantiles. Default: ``10000``
        quantiles (``torch.Tensor`` or tuple or list, optional):
            Specifies a sequence of quantiles to compute, which must
            be between 0 and 1 inclusive. Default:
            ``(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)``
    """

    def add(self, data: Tensor) -> None:
        r"""Adds new data to the summary.

        Args:
            data (``torch.Tensor`` of type float): Specifies the data
                to add to the summary. This method converts the input
                to a ``torch.Tensor`` of type float if the tensor type
                is different.
        """
        if data.numel() > 0:
            values = data.float().flatten()
            self._sum += float(values.sum())
            self._count += values.numel()
            self._min_value = min(self._min_value, values.min().item())
            self._max_value = max(self._max_value, values.max().item())
            self._values.extend(values.tolist())


class FloatTensorSequenceDataSummary(BaseSequenceDataSummary[Tensor]):
    r"""Implements a data summary for ``torch.Tensor``s of type float.

    The input should have a shape ``(sequence_length, *)`` where `*`
    means any number of dimensions. This data summary assumes that
    the data are continuous numerical values. This data summary
    computes the following statistics:

        - ``count``: the number of values
        - ``sum``: the sum of tne values
        - ``mean``: the mean of tne values
        - ``median``: the median of tne values
        - ``std``: the standard deviation of tne values
        - ``max``: the max value
        - ``min``: the min value
        - ``quantiles``: the quantile values

    Args:
        max_size (int, optional): Specifies the maximum size used to
            store the last values because it may not be possible to
            store all the values. This parameter is used to compute
            the median and the quantiles. Default: ``10000``
        quantiles (``torch.Tensor`` or tuple or list, optional):
            Specifies a sequence of quantiles to compute, which must
            be between 0 and 1 inclusive. Default:
            ``(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)``
    """

    def __init__(
        self,
        max_size: int = 10000,
        quantiles: Union[Tensor, tuple[float, ...], list[float]] = DEFAULT_QUANTILES,
    ):
        super().__init__(
            value_summary=FloatTensorDataSummary(max_size=max_size, quantiles=quantiles)
        )

    def _get_sequence_length(self, data: Tensor) -> int:
        r"""Gets the sequence length of the data.

        Args:
            data (``torch.Tensor`` of type float and shape
                ``(sequence_length, *)`` where `*` means any number of
                dimensions): Specifies the input sequence.

        Returns:
            int: The sequence length.
        """
        return data.shape[0]



r"""This module implements some utility functions to manipulate
mappings/dicts."""

__all__ = ["convert_to_dict_of_lists", "convert_to_list_of_dicts", "get_first_value"]

from collections.abc import Hashable, Mapping, Sequence
from typing import Any


def convert_to_dict_of_lists(seq_of_mappings: Sequence[Mapping]) -> dict[Hashable, list]:
    r"""Convert a sequence of mappings to a dictionary of lists.

    All the dictionaries should have the same keys. The first
    dictionary in the sequence is used to find the keys.

    Args:
        seq_of_mappings (sequence): Specifies the sequence of
            mappings.

    Returns:
        dict: A dictionary of lists.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.mapping import convert_to_dict_of_lists
        >>> convert_to_dict_of_lists(
        ...     [{"key1": 1, "key2": 10}, {"key1": 2, "key2": 20}, {"key1": 3, "key2": 30}]
        ... )
        {'key1': [1, 2, 3], 'key2': [10, 20, 30]}
    """
    if seq_of_mappings:
        return {key: [dic[key] for dic in seq_of_mappings] for key in seq_of_mappings[0]}
    return {}


def convert_to_list_of_dicts(mapping_of_seqs: Mapping[Hashable, Sequence]) -> list[dict]:
    r"""Convert a mapping of sequences to a list of dictionaries.

    All the sequences should have the same length.

    Args:
        mapping_of_seqs (``Mapping``): Specifies a mapping of
            sequences.

    Returns:
        dict: A dictionary of lists.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.mapping import convert_to_list_of_dicts
        >>> convert_to_list_of_dicts({"key1": [1, 2, 3], "key2": [10, 20, 30]})
        [{'key1': 1, 'key2': 10}, {'key1': 2, 'key2': 20}, {'key1': 3, 'key2': 30}]
    """
    return [dict(zip(mapping_of_seqs, seqs)) for seqs in zip(*mapping_of_seqs.values())]


def remove_keys_starting_with(mapping: Mapping, prefix: str) -> dict:
    r"""Removes the keys that start with a given prefix.

    Args:
        mapping (``Mapping``): Specifies the original mapping.

    Returns:
        dict: A new dict without the removed keys.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.mapping import remove_keys_starting_with
        >>> remove_keys_starting_with(
        ...     {"key": 1, "key.abc": 2, "abc": 3, "abc.key": 4, 1: 5, (2, 3): 6}, "key",
        ... )
        {'abc': 3, 'abc.key': 4, 1: 5, (2, 3): 6}
    """
    new_dict = {}
    for key, value in mapping.items():
        if isinstance(key, str) and key.startswith(prefix):
            continue
        new_dict[key] = value
    return new_dict


def get_first_value(data: Mapping) -> Any:
    r"""Gets the first value of a mapping.

    Args:
        data (``Mapping``): Specifies the input mapping.

    Returns:
        The first value in the mapping.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.mapping import get_first_value
        >>> get_first_value({"key1": 1, "key2": 2})
        1
    """
    if not data:
        raise ValueError("First value cannot be returned because the mapping is empty")
    return data[next(iter(data))]



r"""This module defines the optimizer base class."""

__all__ = ["setup_optimizer"]

import logging
from typing import Optional, Union

from objectory import factory
from torch.nn import Module
from torch.optim import Optimizer

from gravitorch.nn.utils.module_helpers import has_learnable_parameters
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def setup_optimizer(model: Module, optimizer: Union[Optimizer, dict, None]) -> Optional[Optimizer]:
    r"""Sets up the optimizer.

    The optimizer is instantiated from its configuration by using the
    ``OptimizerFactory`` factory function.

    Args:
        model (``torch.nn.Module``): Specifies the model to train.
        optimizer (``torch.optim.Optimizer`` or dict or ``None``):
            Specifies the optimizer or its configuration. If ``None``,
            no optimizer is instantiated.

    Returns:
        ``torch.optim.Optimizer`` or ``None``: The (instantiated)
            optimizer if the input is not ``None``, otherwise ``None``.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.optimizers import setup_optimizer
        >>> import torch
        >>> optimizer = setup_optimizer(
        ...     torch.nn.Linear(4, 6),
        ...     optimizer={'_target_': "torch.optim.SGD", "lr": 0.01}
        ... )
        >>> optimizer
        SGD (
        Parameter Group 0
            dampening: 0
            foreach: None
            lr: 0.01
            maximize: False
            momentum: 0
            nesterov: False
            weight_decay: 0
        )
        >>> model = torch.nn.Linear(4, 6)
        >>> optimizer = torch.optim.SGD(model.parameters(), lr=0.001)
        >>> setup_optimizer(model, optimizer)
        SGD (
        Parameter Group 0
            dampening: 0
            foreach: None
            lr: 0.001
            maximize: False
            momentum: 0
            nesterov: False
            weight_decay: 0
        )
    """
    if optimizer is None:
        logger.info("No optimizer")
        return None
    if not has_learnable_parameters(model):
        logger.info(
            "The optimizer is not initialized because the model "
            "does not have learnable parameters"
        )
        return None
    if isinstance(optimizer, dict):
        logger.info(
            f"Initializing an optimizer from its configuration... {str_target_object(optimizer)}"
        )
        optimizer = factory(params=model.parameters(), **optimizer)
    else:
        logger.info(
            "The optimizer is already created. The optimizer has to be initialized "
            "with the model parameters."
        )
    return optimizer



__all__ = [
    "BaseEventHandler",
    "BaseEventHandlerWithArguments",
    "ConditionalEventHandler",
    "VanillaEventHandler",
]

from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any, Optional


class BaseEventHandler(ABC):
    r"""Defines the base class to implement an event handler.

    A child class has to implement the following methods:

        - ``__eq__``
        - ``handle``
    """

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        r"""Compares two event handlers."""

    @abstractmethod
    def handle(self) -> None:
        r"""Handles the event."""


class BaseEventHandlerWithArguments(BaseEventHandler):
    r"""Defines a base class to implement an event handler with positional
    and/or keyword arguments.

    A child class has to implement the ``__eq__`` method.

    Args:
        handler (callable): Specifies the handler.
        handler_args (tuple or ``None``, optional): Specifies the
            positional argument of the handler. Default: ``None``
        handler_kwargs (dict or ``None``, optional): Specifies the
            arbitrary keyword arguments of the handler.
            Default: ``None``
    """

    def __init__(
        self,
        handler: Callable,
        handler_args: Optional[tuple] = None,
        handler_kwargs: Optional[dict] = None,
    ):
        self._handler = handler
        self._handler_args = handler_args or tuple()
        self._handler_kwargs = handler_kwargs or {}

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(handler={self._handler}, "
            f"handler_args={self._handler_args}, handler_kwargs={self._handler_kwargs})"
        )

    @property
    def handler(self) -> Callable:
        r"""callable: The handler."""
        return self._handler

    @property
    def handler_args(self) -> tuple:
        r"""tuple: Variable length argument list of the handler."""
        return self._handler_args

    @property
    def handler_kwargs(self) -> dict:
        r"""dict: Arbitrary keyword arguments of the handler."""
        return self._handler_kwargs

    def handle(self) -> None:
        r"""Handles the event."""
        self._handler(*self._handler_args, **self._handler_kwargs)


class VanillaEventHandler(BaseEventHandlerWithArguments):
    r"""Implements a simple event handler."""

    def __eq__(self, other: Any) -> bool:
        r"""Compares two event handlers."""
        if isinstance(other, VanillaEventHandler):
            return (
                self.handler == other.handler
                and self.handler_args == other.handler_args
                and self.handler_kwargs == other.handler_kwargs
            )
        return False


class ConditionalEventHandler(BaseEventHandlerWithArguments):
    r"""Implements a conditional event handler.

    The handler is executed only if the condition is ``True``.

    Args:
        handler (callable): Specifies the handler.
        condition (callable): Specifies the condition for this event
            handler. The condition should be callable without
            arguments.
        handler_args (tuple or ``None``): Specifies the positional
            argument of the handler.
        handler_kwargs (dict): Specifies the arbitrary keyword
            arguments of the handler.
    """

    def __init__(
        self,
        handler: Callable,
        condition: Callable,
        handler_args: Optional[tuple] = None,
        handler_kwargs: Optional[dict] = None,
    ):
        super().__init__(handler=handler, handler_args=handler_args, handler_kwargs=handler_kwargs)
        if not callable(condition):
            raise ValueError(f"The condition is not callable (received: {condition})")
        self._condition = condition

    def __eq__(self, other: Any) -> bool:
        r"""Compares two event handlers."""
        if isinstance(other, ConditionalEventHandler):
            return (
                self.handler == other.handler
                and self.handler_args == other.handler_args
                and self.handler_kwargs == other.handler_kwargs
                and self.condition == other.condition
            )
        return False

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(handler={self._handler}, "
            f"handler_args={self._handler_args}, handler_kwargs={self._handler_kwargs}, "
            f"condition={self._condition})"
        )

    @property
    def condition(self) -> Callable:
        r"""callable: The condition."""
        return self._condition

    def handle(self) -> None:
        r"""Handles the event."""
        if self._condition():
            super().handle()



__all__ = ["setup_data_source", "setup_and_attach_data_source"]

import logging
from typing import Union

from gravitorch.datasources.base import BaseDataSource
from gravitorch.engines.base import BaseEngine
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def setup_data_source(data_source: Union[BaseDataSource, dict]) -> BaseDataSource:
    r"""Sets up a data source.

    The data source is instantiated from its configuration by using
    the ``BaseDataSource`` factory function.

    Args:
        data_source (``BaseDataSource`` or dict): Specifies the data
            source or its configuration.

    Returns:
        ``BaseDataSource``: The instantiated data source.
    """
    if isinstance(data_source, dict):
        logger.info(
            "Initializing a data source from its configuration... "
            f"{str_target_object(data_source)}"
        )
        data_source = BaseDataSource.factory(**data_source)
    return data_source


def setup_and_attach_data_source(
    data_source: Union[BaseDataSource, dict], engine: BaseEngine
) -> BaseDataSource:
    r"""Sets up a data source and attach it to an engine.

    Note that if you call this function ``N`` times with the same data
    source object, the data source will be attached ``N`` times to the
    engine.

    Args:
        data_source (``BaseDataSource`` or dict): Specifies the data
            source or its configuration.
        engine (``BaseEngine``): Specifies the engine.

    Returns:
        ``BaseDataSource``: The instantiated data source.
    """
    data_source = setup_data_source(data_source)
    logger.info("Adding a data source object to an engine...")
    data_source.attach(engine)
    return data_source



r"""This module implements a handler to load state dict."""

__all__ = ["ModelStateDictLoader", "PartialModelStateDictLoader"]

import logging
from collections.abc import Sequence
from pathlib import Path
from typing import Optional, Union

from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.nn.utils.state_dict import (
    load_checkpoint_to_module,
    load_model_state_dict,
)
from gravitorch.utils.events import VanillaEventHandler
from gravitorch.utils.path import sanitize_path

logger = logging.getLogger(__name__)


class ModelStateDictLoader(BaseHandler):
    r"""Implements a handler to load the model state dict.

    Args:
        checkpoint_path (``pathlib.Path`` or str): Specifies a path
            to a model checkpoint. This weights in the checkpoint are
            used to initialize the model.
        event (str, optional): Specifies the event when to load the
            model state dict. Default: ``'started'``
        strict (bool, optional): whether to strictly enforce that the
            keys in ``state_dict`` match the keys returned by this
            module's :meth:`~torch.nn.Module.state_dict` function.
            Default: ``True``
        key (str or list or tuple or ``None``, optional): Specifies
            the key of the state dict to load. The state dict can
            contain data that are not about the module, so they may
            need to be excluded. For nested case, it is possible to
            specify the list of keys to get the good part of the state
            dict. Default: ``None``
    """

    def __init__(
        self,
        checkpoint_path: Union[Path, str],
        event: str = EngineEvents.STARTED,
        strict: bool = True,
        key: Union[str, list[str], tuple[str, ...], None] = None,
    ):
        self._checkpoint_path = sanitize_path(checkpoint_path)
        self._event = str(event)
        self._strict = bool(strict)
        self._key = key

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(checkpoint_path={self._checkpoint_path}, "
            f"event={self._event}, strict={self._strict}, key={self._key})"
        )

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=VanillaEventHandler(
                self.load,
                handler_kwargs={"engine": engine},
            ),
        )

    def load(self, engine: BaseEngine) -> None:
        r"""Loads a model state dict.

        Args:
            engine (``BaseEngine``): Specifies the engine with the
                model.
        """
        load_checkpoint_to_module(
            path=self._checkpoint_path,
            module=engine.model,
            strict=self._strict,
            key=self._key,
        )


class PartialModelStateDictLoader(BaseHandler):
    r"""Implements a handler to load some model weights from a checkpoint.

    Args:
        checkpoint_path (``pathlib.Path`` or str): Specifies a path
            to a model checkpoint. This weights in the checkpoint are
            used to initialize the model.
        event (str, optional): Specifies the event when to load the
            model state dict. Default: ``'started'``
        strict (bool, optional): whether to strictly enforce that the
            keys in ``state_dict`` match the keys returned by this
            module's :meth:`~torch.nn.Module.state_dict` function.
            Default: ``True``
        exclude_key_prefixes (sequence or ``None``, optional):
            Specifies the list of key prefixes to exclude when loading
            the state dict. Default: ``None``
    """

    def __init__(
        self,
        checkpoint_path: Union[Path, str],
        event: str = EngineEvents.STARTED,
        strict: bool = True,
        exclude_key_prefixes: Optional[Sequence[str]] = None,
    ):
        self._checkpoint_path = sanitize_path(checkpoint_path)
        self._event = str(event)
        self._strict = bool(strict)
        self._exclude_key_prefixes = exclude_key_prefixes or []

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(checkpoint_path={self._checkpoint_path}, event={self._event}, "
            f"strict={self._strict}, exclude_key_prefixes={self._exclude_key_prefixes})"
        )

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=VanillaEventHandler(
                self.load,
                handler_kwargs={"engine": engine},
            ),
        )

    def load(self, engine: BaseEngine) -> None:
        r"""Loads a model state dict.

        Args:
            engine (``BaseEngine``): Specifies the engine with the
                model.
        """
        load_model_state_dict(
            path=self._checkpoint_path,
            module=engine.model,
            strict=self._strict,
            exclude_key_prefixes=self._exclude_key_prefixes,
        )



r"""This module implements some conditions used in the evaluation loops."""

__all__ = ["BaseEvalCondition", "EveryEpochEvalCondition", "LastEpochEvalCondition"]

from abc import abstractmethod

from objectory import AbstractFactory

from gravitorch.engines.base import BaseEngine


class BaseEvalCondition(metaclass=AbstractFactory):
    r"""Defines the base class for the conditions of the evaluation loop.

    These conditions indicate if the evaluation loop should be executed
    or not.
    """

    @abstractmethod
    def __call__(self, engine: BaseEngine) -> bool:
        r"""Indicates if the evaluation loop should be executed.

        Args:
            engine (``BaseEngine``): Specifies the engine.

        Returns:
            bool: ``True`` if the evaluation loop should be evaluated,
                otherwise ``False``
        """


class EveryEpochEvalCondition(BaseEvalCondition):
    r"""Implements a condition that is true every N epoch.

    Args:
        every (int, optional): Specifies the frequency of the
            evaluation. Default: ``1``
    """

    def __init__(self, every: int = 1):
        self._every = every

    def __call__(self, engine: BaseEngine) -> bool:
        r"""Indicates if it is a multiple of ``every``.

        Args:
            engine (``BaseEngine``): Specifies the engine.

        Returns:
            bool: ``True`` if the evaluation loop should be evaluated,
                otherwise ``False``
        """
        return engine.epoch % self._every == 0

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(every={self._every})"


class LastEpochEvalCondition(BaseEvalCondition):
    r"""Implements a condition that is true only for the last epoch."""

    def __call__(self, engine: BaseEngine) -> bool:
        r"""Indicates if it is the last epoch.

        Args:
            engine (``BaseEngine``): Specifies the engine.

        Returns:
            bool: ``True`` if the evaluation loop should be evaluated,
                otherwise ``False``
        """
        return engine.epoch == engine.max_epochs - 1

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"



__all__ = ["app_vae_time_loss", "exponential_nll_loss"]

from gravitorch.nn.functional.experimental.app_vae import app_vae_time_loss
from gravitorch.nn.functional.experimental.exponential_nll import exponential_nll_loss



r"""This module implements some utility functions for the profilers."""

__all__ = ["setup_profiler"]

import logging
from typing import Union

from gravitorch.utils.format import str_target_object
from gravitorch.utils.profilers import BaseProfiler, NoOpProfiler

logger = logging.getLogger(__name__)


def setup_profiler(profiler: Union[BaseProfiler, dict, None]) -> BaseProfiler:
    r"""Sets up the profiler.

    The profiler is instantiated from its configuration by using the
    ``BaseProfiler`` factory function.

    Args:
        profiler (``BaseProfiler`` or dict or None): Specifies the
            profiler or its configuration. If ``None``, the
            ``NoOpProfiler`` is instantiated.

    Returns:
        ``BaseProfiler``: A profiler.
    """
    if profiler is None:
        profiler = NoOpProfiler()
    if isinstance(profiler, dict):
        logger.info(
            f"Initializing a profiler from its configuration... {str_target_object(profiler)}"
        )
        profiler = BaseProfiler.factory(**profiler)
    return profiler



r"""This module implements some utility functions for the engine states."""

__all__ = ["setup_engine_state"]

import logging
from typing import Union

from gravitorch.utils.engine_states.base import BaseEngineState
from gravitorch.utils.engine_states.vanilla import VanillaEngineState
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def setup_engine_state(state: Union[BaseEngineState, dict, None]) -> BaseEngineState:
    r"""Sets up the engine state.

    The state is instantiated from its configuration by using the
    ``BaseEngineState`` factory function.

    Args:
        state (``BaseEngineState`` or dict or None): Specifies the
            engine state or its configuration. If ``None``, the
            ``VanillaEngineState`` is instantiated.

    Returns:
        ``BaseEngineState``: The engine state.
    """
    if state is None:
        state = VanillaEngineState()
    if isinstance(state, dict):
        logger.info(
            f"Initializing an engine state from its configuration... {str_target_object(state)}"
        )
        state = BaseEngineState.factory(**state)
    return state



__all__ = ["EarlyStopping"]

import logging
import operator
from collections.abc import Callable

from gravitorch import constants as ct
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.utils.events import VanillaEventHandler
from gravitorch.utils.history import BaseHistory, MaxScalarHistory, MinScalarHistory

logger = logging.getLogger(__name__)


class EarlyStopping(BaseHandler):
    r"""Implements an early stopping handler to stop the training if no
    improvement after a given number of epochs.

    This early stopping handler only works with
    ``MaxScalarHistory`` and ``MinScalarHistory``
    metrics.

    Args:
        metric_name (str, optional): Specifies the metric name that
            is used to measure the improvement. By default, the metric
            is the loss on the evaluation data set.
            Default: 'eval/loss'.
        patience (int, optional): Specifies the number of epochs to
            wait if no improvement and then stop the training.
            Default: ``5``
        delta (float, optional): A minimum increase or decrease in
            the score to qualify as an improvement, i.e. an increase
            of less than or equal to `delta`, will count as no
            improvement. Default: ``0.0``
        cumulative_delta (bool, optional): It True, `min_delta`
            defines an increase since the last `patience` reset,
            otherwise, it defines an increase after the last epoch.
            Default: ``False``.

    Example usage:

    .. code-block:: python

        # Create an engine
        >>> engine = ...
        >>> from gravitorch.handlers import EarlyStopping
        # Add the early stopping handler to the engine with a patience of 10 epochs
        # for the accuracy metric.
        >>> handler = EarlyStopping(metric_name='eval/accuracy', patience=10)
        >>> handler.attach(engine)
    """

    def __init__(
        self,
        metric_name: str = f"{ct.EVAL}/loss",
        patience: int = 5,
        delta: float = 0.0,
        cumulative_delta: bool = False,
    ):
        self._metric_name = str(metric_name)
        if patience < 1:
            raise ValueError(f"Argument patience should be positive integer (received: {patience})")
        self._patience = int(patience)
        if delta < 0.0:
            raise ValueError(
                f"Argument 'min_delta' should not be a negative number  (received: {patience})"
            )
        self._delta = float(delta)
        self._cumulative_delta = bool(cumulative_delta)

        self._waiting_counter = 0
        self._best_score = None
        self._best_epoch = None

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(metric_name={self._metric_name}, "
            f"patience={self._patience}, "
            f"delta={self._delta}, "
            f"cumulative_delta={self._cumulative_delta}, "
            f"waiting_counter={self._waiting_counter}, "
            f"best_score={self._best_score}, "
            f"best_epoch={self._best_epoch})"
        )

    def attach(self, engine: BaseEngine) -> None:
        logger.info(
            f"Attach early stopping handler for metric {self._metric_name} with the parameters: "
            f"patience={self._patience:,}   delta={self._delta}   "
            f"cumulative_delta={self._cumulative_delta}"
        )
        add_unique_event_handler(
            engine=engine,
            event=EngineEvents.TRAIN_STARTED,
            event_handler=VanillaEventHandler(self.start, handler_kwargs={"engine": engine}),
        )
        add_unique_event_handler(
            engine=engine,
            event=EngineEvents.EPOCH_COMPLETED,
            event_handler=VanillaEventHandler(self.step, handler_kwargs={"engine": engine}),
        )

        # Add the module to the engine so the state of the early stopping module is saved when
        # the engine checkpoint is created. It is necessary to store the state to resume
        # the training.
        engine.add_module(ct.EARLY_STOPPING, self)

        if engine.has_history(self._metric_name):
            self._check_history(engine.get_history(self._metric_name))
        else:
            logger.warning(
                f"There is no history for '{self._metric_name}'. The history needs to be created "
                f"before the `step` method is called"
            )

    def load_state_dict(self, state_dict: dict) -> None:
        r"""Loads the state values from a dict.

        Args:
            state_dict (dict): a dict with parameters

        Example:

        .. code-block:: python

            >>> handler = EarlyStopping(...)
            >>> state = {...}
            >>> handler.load_state_dict(state)
        """
        self._best_epoch = state_dict["best_epoch"]
        self._best_score = state_dict["best_score"]
        self._waiting_counter = state_dict["waiting_counter"]

    def state_dict(self) -> dict:
        r"""Gets a dictionary containing state values.

        Returns:
            dict: the state values in a dict.

        Example:

        .. code-block:: python

            >>> state = EarlyStopping(...)
            >>> state_dict = state.state_dict()
        """
        return {
            "best_epoch": self._best_epoch,
            "best_score": self._best_score,
            "waiting_counter": self._waiting_counter,
        }

    def start(self, engine: BaseEngine) -> None:
        r"""Stops the training if the requirements to stop the training are
        already met.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        logger.info(
            f"Early stopping [start] | {self._metric_name} | "
            f"best_epoch {self._best_epoch} | best_score {self._best_score} | "
            f"waiting_counter {self._waiting_counter}/{self._patience}"
        )
        if self._waiting_counter >= self._patience:
            logger.info(
                f"Early stopping [start] | cancel training because '{self._metric_name}' did not "
                f"improve during the last {self._waiting_counter} epochs"
            )
            engine.terminate()

    def step(self, engine: BaseEngine) -> None:
        """Updates the early stopping handler by using the last value of the
        monitored metric.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        history = engine.get_history(self._metric_name)
        current_score = history.get_last_value()
        min_delta = self._get_min_delta(history)
        comparator = self._get_comparator(history)

        if self._best_score is None:
            # The best score is undefined at the first call.
            # The first score becomes the best score.
            self._best_score = current_score
            self._best_epoch = engine.epoch

        elif not comparator(current_score - min_delta, self._best_score):
            # The current score is not better than the best score + min delta.
            self._waiting_counter += 1
            if not self._cumulative_delta and comparator(current_score, self._best_score):
                self._best_score = current_score
            if self._waiting_counter >= self._patience:
                logger.info(
                    f"Early stopping [step] | stop training because '{self._metric_name}' did not "
                    f"improve during the last {self._waiting_counter:,} epochs"
                )
                engine.terminate()

        else:
            # The current score is better than the best score.
            self._best_score = current_score
            self._waiting_counter = 0
            self._best_epoch = engine.epoch

        logger.info(
            f"Early stopping [step] | {self._metric_name} | "
            f"best_epoch {self._best_epoch} | best_score {self._best_score} | "
            f"waiting_counter {self._waiting_counter}/{self._patience}"
        )

    def _check_history(self, history: BaseHistory) -> None:
        r"""Checks the history.

        Args:
            history (``BaseHistory``): Specifies the history
                tracker to check.

        Raises:
            RuntimeError if the history tracker is not valid.
        """
        if not isinstance(history, (MaxScalarHistory, MinScalarHistory)):
            raise RuntimeError(
                f"The early stopping handler only supports ``MaxScalarHistory`` "
                f"or ``MinScalarHistory`` history tracker (received: {history})"
            )

    def _get_comparator(self, history: BaseHistory) -> Callable:
        r"""Gets the comparator function from the history tracker type.

        Args:
            history (``BaseHistory``): Specifies the history
                tracker.

        Returns:
            callable: The comparator.
        """
        if isinstance(history, MaxScalarHistory):
            return operator.gt  # >
        return operator.lt  # <

    def _get_min_delta(self, history: BaseHistory) -> float:
        r"""Gets the minimum delta.

        Args:
            history (``BaseHistory``): Specifies the history
                tracker.

        Returns:
            float: The minimum delta.
        """
        self._check_history(history)
        return self._delta if isinstance(history, MaxScalarHistory) else -self._delta



__all__ = ["asinh_barron_robust_loss", "barron_robust_loss"]

import math
from typing import Optional

from torch import Tensor

from gravitorch.nn.functional.loss_helpers import basic_loss_reduction


def barron_robust_loss(
    prediction: Tensor,
    target: Tensor,
    alpha: float = 2.0,
    scale: float = 1.0,
    max_value: Optional[float] = None,
    reduction: str = "mean",
) -> Tensor:
    r"""Computes the Barron robust loss.

    Based on the paper:

        A General and Adaptive Robust Loss Function
        Jonathan T. Barron
        CVPR 2019 (https://arxiv.org/abs/1701.03077)

    Args:
        prediction (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the predictions.
        target (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the target values.
        alpha (float, optional): Specifies the shape parameter that
            controls the robustness of the loss. Default: ``2.0``
        scale (float, optional): Specifies the scale parameter that
            controls the size of the loss’s quadratic bowl near 0.
            Default: ``1.0``
        max_value (float or ``None``, optional): Specifies the max
            value to clip the loss before to compute the reduction.
            ``None`` means no clipping is used. Default: ``None``
        reduction (string, optional): Specifies the reduction to
            apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``"mean"``

    Returns:
        ``torch.Tensor`` of type float: The computed loss. The shape
            of the tensor depends on the reduction strategy.
    """
    squared_error = prediction.sub(target).div(scale).pow(2)
    if alpha == 2:
        loss = squared_error
    elif alpha == 0:
        loss = squared_error.mul(0.5).add(1).log()
    else:
        alpha2 = math.fabs(alpha - 2)
        loss = squared_error.div(alpha2).add(1).pow(alpha / 2).sub(1).mul(alpha2 / alpha)
    if max_value is not None:
        loss = loss.clamp(max=max_value)
    return basic_loss_reduction(loss, reduction)


def asinh_barron_robust_loss(
    prediction: Tensor,
    target: Tensor,
    alpha: float = 2.0,
    scale: float = 1.0,
    max_value: Optional[float] = None,
    reduction: str = "mean",
) -> Tensor:
    r"""Computes the Barron loss on the inverse hyperbolic sine (arcsinh)
    transformed prediction and target.

    Args:
        prediction (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the predictions.
        target (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the target values.
        alpha (float, optional): Specifies the shape parameter that
            controls the robustness of the loss. Default: ``2.0``
        scale (float, optional): Specifies the scale parameter that
            controls the size of the loss’s quadratic bowl near 0.
            Default: ``1.0``
        max_value (float or ``None``, optional): Specifies the max
            value to clip the loss before to compute the reduction.
            ``None`` means no clipping is used. Default: ``None``
        reduction (string, optional): Specifies the reduction to
            apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``"mean"``

    Returns:
        ``torch.Tensor`` of type float: The computed loss. The shape
            of the tensor depends on the reduction strategy.
    """
    return barron_robust_loss(
        prediction=prediction.asinh(),
        target=target.asinh(),
        alpha=alpha,
        scale=scale,
        max_value=max_value,
        reduction=reduction,
    )



r"""This package contains the code to initialize the model parameters."""

__all__ = [
    "BaseDefaultParameterInitializer",
    "BaseKaimingParameterInitializer",
    "BaseParameterInitializer",
    "BaseXavierParameterInitializer",
    "ConstantBiasParameterInitializer",
    "KaimingNormalParameterInitializer",
    "KaimingUniformParameterInitializer",
    "NoParameterInitializer",
    "SequentialParameterInitializer",
    "TruncNormalParameterInitializer",
    "XavierNormalParameterInitializer",
    "XavierUniformParameterInitializer",
    "bias_constant_",
    "recursive_bias_constant_",
    "recursive_constant_",
    "recursive_kaiming_normal_",
    "recursive_kaiming_uniform_",
    "recursive_trunc_normal_",
    "recursive_xavier_normal_",
    "recursive_xavier_uniform_",
    "setup_parameter_initializer",
]

from gravitorch.utils.parameter_initializers.base import (
    BaseDefaultParameterInitializer,
    BaseParameterInitializer,
)
from gravitorch.utils.parameter_initializers.constant import (
    ConstantBiasParameterInitializer,
    bias_constant_,
    recursive_bias_constant_,
    recursive_constant_,
)
from gravitorch.utils.parameter_initializers.kaiming import (
    BaseKaimingParameterInitializer,
    KaimingNormalParameterInitializer,
    KaimingUniformParameterInitializer,
    recursive_kaiming_normal_,
    recursive_kaiming_uniform_,
)
from gravitorch.utils.parameter_initializers.no_init import NoParameterInitializer
from gravitorch.utils.parameter_initializers.sequential import (
    SequentialParameterInitializer,
)
from gravitorch.utils.parameter_initializers.truncated_normal import (
    TruncNormalParameterInitializer,
    recursive_trunc_normal_,
)
from gravitorch.utils.parameter_initializers.utils import setup_parameter_initializer
from gravitorch.utils.parameter_initializers.xavier import (
    BaseXavierParameterInitializer,
    XavierNormalParameterInitializer,
    XavierUniformParameterInitializer,
    recursive_xavier_normal_,
    recursive_xavier_uniform_,
)



r"""This module implements some utility functions to use ``torch.Tensor``s."""

__all__ = ["isymlog", "isymlog_", "safeexp", "safelog", "symlog", "symlog_", "scalable_quantile"]

import numpy as np
import torch
from torch import Tensor


def scalable_quantile(tensor: Tensor, q: Tensor, method: str = "linear") -> Tensor:
    r"""Implements a scalable function to compute the ``q``-th quantiles.

    Note: this function uses numpy to compute the ``q``-th quantiles
    because PyTorch has a limit to 16M items.
    https://github.com/pytorch/pytorch/issues/64947

    Args:
        tensor (``torch.Tensor``): Specifies the tensor of values.
        q (``torch.Tensor`` of type float and shape
            ``(num_q_values,)``): Specifies the ``q``-values in
            the range ``[0, 1]``.
        method (str, optional): Specifies the interpolation
            method to use when the desired quantile lies between
            two data points. Can be ``'linear'``, ``'lower'``,
            ``'higher'``, ``'midpoint'`` and ``'nearest'``.
            Default: ``'linear'``.

    Returns:
        ``torch.Tensor`` of shape  ``(num_q_values,)`` and type float:
            The ``q``-th quantiles.
    """
    return torch.from_numpy(
        np.quantile(tensor.detach().cpu().numpy(), q=q.numpy(), method=method)
    ).float()


def safeexp(tensor: Tensor, max_value: float = 20.0) -> Tensor:
    r"""Computes safely the exponential of the elements.

    The values that are higher than the specified minimum value are
    set to this maximum value. Using a not too large positive value
    leads to an output tensor without Inf.

    Args:
        tensor (``torch.Tensor``): Specifies the input tensor.
        max_value (float, optional): Specifies the maximum value.
            Default: ``20.0``

    Returns:
        ``torch.Tensor``: A new tensor with the exponential of the
            elements.
    """
    return tensor.clamp(max=max_value).exp()


def safelog(tensor: Tensor, min_value: float = 1e-8) -> Tensor:
    r"""Computes safely the logarithm natural logarithm of the elements.

    The values that are lower than the specified minimum value are set
    to this minimum value. Using a small positive value leads to an
    output tensor without NaN or Inf.

    Args:
        tensor (``torch.Tensor``): Specifies the input tensor.
        min_value (float, optional): Specifies the minimum value.
            Default: ``1e-8``

    Returns:
        ``torch.Tensor``: A new tensor with the natural logarithm
            of the elements.
    """
    return tensor.clamp(min=min_value).log()


def symlog(tensor: Tensor) -> Tensor:
    r"""Computes the symmetric logarithm natural logarithm of the elements.

    Note this transformation supports negative values.

    Args:
        tensor (``torch.Tensor``): Specifies the input tensor.

    Returns:
        ``torch.Tensor``: A new tensor with the symmetric natural
            logarithm of the elements.
    """
    return tensor.sign().mul(tensor.abs().log1p())


def symlog_(tensor: Tensor) -> None:
    r"""Computes the symmetric logarithm natural of the elements.

    In-place version of ``symlog``.

    Note this transformation supports negative values.

    Args:
        tensor (``torch.Tensor``): Specifies the input tensor.
    """
    sign = tensor.sign()
    tensor.abs_()
    tensor.log1p_()
    tensor.mul_(sign)


def isymlog(tensor: Tensor, max_value: float = 10) -> Tensor:
    r"""Computes the inverse of the symmetric logarithm natural of the elements.

    Args:
        tensor (``torch.Tensor``): Specifies the input tensor.

    Returns:
        ``torch.Tensor``: A new tensor with the inverse of the
            symmetric natural logarithm of the elements.
    """
    sign = tensor.sign()
    return sign.mul(tensor.mul(sign).clamp(max=max_value).exp().sub(1))


def isymlog_(tensor: Tensor, max_value: float = 10) -> None:
    r"""Computes the inverse of the symmetric logarithm natural of the elements.

    In-place version of ``isymlog``.

    Args:
        tensor (``torch.Tensor``): Specifies the input tensor.
    """
    sign = tensor.sign()
    tensor.mul_(sign)
    tensor.clamp_(max=max_value)
    tensor.exp_()
    tensor.sub_(1)
    tensor.mul_(sign)



r"""This module implements some sum-like fusion layers."""

__all__ = ["SumFusion", "AverageFusion"]

from torch import Tensor
from torch.nn import Module


class SumFusion(Module):
    r"""Defines a layer to sum the inputs.

    Args:
        normalized (bool, optional): Specifies the output is
            normalized by the number of inputs.
            Default: ``False``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn import SumFusion
        >>> module = SumFusion()
        >>> module
        SumFusion(normalized=False)
        >>> x1 = torch.tensor([[2, 3, 4], [5, 6, 7]], dtype=torch.float, requires_grad=True)
        >>> x2 = torch.tensor([[12, 13, 14], [15, 16, 17]], dtype=torch.float, requires_grad=True)
        >>> out = module(x1, x2)
        >>> out
        tensor([[14., 16., 18.],
                [20., 22., 24.]], grad_fn=<AddBackward0>)
        >>> out.mean().backward()
    """

    def __init__(self, normalized: bool = False):
        super().__init__()
        self._normalized = normalized

    def extra_repr(self) -> str:
        return f"normalized={self._normalized}"

    def forward(self, *inputs: Tensor) -> Tensor:
        r"""Sums the list or tuple of inputs.

        Args:
            *inputs (list or tuple of tensors): Specifies the list or
                tuple of tensors to sum. The shape of the tensors
                should be the same. By default, this layer expects
                that each input is a ``torch.Tensor`` of shape
                ``(batch size, feature size)``. But it can also work
                if the inputs have a shape
                ``(sequence length, batch size, feature size)`` or
                similar shapes.

        Returns:
            ``torch.Tensor`` with the same shape that the input
                tensor: The fused tensor.
        """
        if not inputs:
            raise ValueError("This fusion needs to have at least one tensor in the input")

        output = inputs[0]
        for x in inputs[1:]:
            output = output + x

        if self._normalized:
            output = output.div(len(inputs))
        return output


class AverageFusion(SumFusion):
    r"""Implements a layer to average the inputs.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn import AverageFusion
        >>> module = AverageFusion()
        >>> module
        AverageFusion(normalized=True)
        >>> x1 = torch.tensor([[2, 3, 4], [5, 6, 7]], dtype=torch.float, requires_grad=True)
        >>> x2 = torch.tensor([[12, 13, 14], [15, 16, 17]], dtype=torch.float, requires_grad=True)
        >>> out = module(x1, x2)
        >>> out
        tensor([[ 7.,  8.,  9.],
                [10., 11., 12.]], grad_fn=<DivBackward0>)
        >>> out.mean().backward()
    """

    def __init__(self):
        super().__init__(normalized=True)



__all__ = ["setup_nn_module"]

import logging
from typing import Union

from objectory import factory
from torch.nn import Module

from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def setup_nn_module(module: Union[Module, dict]) -> Module:
    r"""Sets up a ``torch.nn.Module`` object.

    Args:
        module (``torch.nn.Module`` or dict): Specifies the module or
            its configuration (dictionary).

    Returns:
        ``torch.nn.Module``: The instantiated ``torch.nn.Module``
            object.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.nn import setup_nn_module
        >>> linear = setup_nn_module(
        ...     {'_target_': 'torch.nn.Linear', 'in_features': 4, 'out_features': 6}
        ... )
        >>> linear
        Linear(in_features=4, out_features=6, bias=True)
        >>> setup_nn_module(linear)  # Do nothing because the module is already instantiated
        Linear(in_features=4, out_features=6, bias=True)
    """
    if isinstance(module, dict):
        logger.info(
            "Initializing a `torch.nn.Module` from its configuration... "
            f"{str_target_object(module)}"
        )
        module = factory(**module)
    return module



r"""This module defines the base comparator and some implementations."""

__all__ = [
    "BaseComparator",
    "ComparatorAllCloseOperator",
    "ComparatorEqualityOperator",
    "MaxScalarComparator",
    "MinScalarComparator",
]

import logging
from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar, Union

from coola import (
    AllCloseTester,
    BaseAllCloseOperator,
    BaseAllCloseTester,
    BaseEqualityOperator,
    BaseEqualityTester,
    EqualityTester,
)

T = TypeVar("T")

logger = logging.getLogger(__name__)


class BaseComparator(Generic[T], ABC):
    r"""Definition of the base comparator class."""

    @abstractmethod
    def equal(self, other: Any) -> bool:
        r"""Indicates if two comparators are equal or not.

        Args:
            other: Specifies the value to compare.

        Returns:
            bool: ``True`` if the comparators are equal,
                ``False`` otherwise.
        """

    @abstractmethod
    def get_initial_best_value(self) -> T:
        r"""Gets the initial best value.

        Returns:
            The initial best value.
        """

    @abstractmethod
    def is_better(self, old_value: T, new_value: T) -> bool:
        r"""Indicates if the new value is better than the old value.

        Args:
            old_value: Specifies the old value to compare.
            new_value: Specifies the new value to compare.

        Returns:
            bool: ``True`` if the new value is better than the old
                value, otherwise ``False``.
        """


class MaxScalarComparator(BaseComparator[Union[float, int]]):
    r"""Implementation of a max comparator for scalar value.

    This comparator can be used to find the maximum value between two
    scalar values.
    """

    def equal(self, other: Any) -> bool:
        return isinstance(other, MaxScalarComparator)

    def get_initial_best_value(self) -> float:
        return -float("inf")

    def is_better(self, old_value: Union[float, int], new_value: Union[float, int]) -> bool:
        return old_value <= new_value


class MinScalarComparator(BaseComparator[Union[float, int]]):
    r"""Implementation of a min comparator for scalar value.

    This comparator can be used to find the minimum value between two
    scalar values.
    """

    def equal(self, other: Any) -> bool:
        return isinstance(other, MinScalarComparator)

    def get_initial_best_value(self) -> float:
        return float("inf")

    def is_better(self, old_value: Union[float, int], new_value: Union[float, int]) -> bool:
        return new_value <= old_value


class ComparatorAllCloseOperator(BaseAllCloseOperator[BaseComparator]):
    r"""Implements an allclose operator for ``BaseComparator`` objects."""

    def allclose(
        self,
        tester: BaseAllCloseTester,
        object1: BaseComparator,
        object2: Any,
        rtol: float = 1e-5,
        atol: float = 1e-8,
        equal_nan: bool = False,
        show_difference: bool = False,
    ) -> bool:
        if not isinstance(object2, BaseComparator):
            if show_difference:
                logger.info(f"object2 is not a `BaseComparator` object: {type(object2)}")
            return False
        object_equal = object1.equal(object2)
        if show_difference and not object_equal:
            logger.info(
                f"`BaseComparator` objects are different\nobject1=\n{object1}\nobject2=\n{object2}"
            )
        return object_equal


class ComparatorEqualityOperator(BaseEqualityOperator[BaseComparator]):
    r"""Implements an equality operator for ``BaseComparator`` objects."""

    def equal(
        self,
        tester: BaseEqualityTester,
        object1: BaseComparator,
        object2: Any,
        show_difference: bool = False,
    ) -> bool:
        if not isinstance(object2, BaseComparator):
            if show_difference:
                logger.info(f"object2 is not a `BaseComparator` object: {type(object2)}")
            return False
        object_equal = object1.equal(object2)
        if show_difference and not object_equal:
            logger.info(
                f"`BaseComparator` objects are different\nobject1=\n{object1}\nobject2=\n{object2}"
            )
        return object_equal


if not AllCloseTester.has_allclose_operator(BaseComparator):
    AllCloseTester.add_allclose_operator(
        BaseComparator, ComparatorAllCloseOperator()
    )  # pragma: no cover
if not EqualityTester.has_equality_operator(BaseComparator):
    EqualityTester.add_equality_operator(
        BaseComparator, ComparatorEqualityOperator()
    )  # pragma: no cover



__all__ = ["NoOpDevicePlacement"]

from typing import Any

from gravitorch.utils.device_placement.base import BaseDevicePlacement


class NoOpDevicePlacement(BaseDevicePlacement):
    r"""Implements a no-operation device placement.

    This class does not change the device placement of the object or in
    other words, the target device is always the device where the object
    is.
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def send(self, obj: Any) -> Any:
        r"""Sends the object on a target device.

        Args:
            obj: The object to send to the target device.

        Returns:
            The object on the target device.
        """
        return obj



r"""This module defines some conditions that can be used in the event
system."""
from __future__ import annotations

__all__ = ["EpochPeriodicCondition", "IterationPeriodicCondition", "PeriodicCondition"]

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine


class PeriodicCondition:
    r"""Implements a periodic condition.

    This condition is true every ``freq`` events.

    Args:
        freq (int): Specifies the frequency.
    """

    def __init__(self, freq: int):
        self._freq = int(freq)
        self._step = 0

    def __call__(self) -> bool:
        r"""Evaluates the condition given the current state.

        Returns:
            bool: ``True`` if the condition is ``True``, otherwise
                ``False``.
        """
        condition = self._step % self._freq == 0
        self._step += 1
        return condition

    def __eq__(self, other) -> bool:
        if isinstance(other, PeriodicCondition):
            return self.freq == other.freq
        return False

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(freq={self._freq:,}, step={self._step:,})"

    @property
    def freq(self) -> int:
        r"""int: The frequency of the condition."""
        return self._freq


class EpochPeriodicCondition:
    r"""Implements an epoch periodic condition.

    This condition is true every ``freq`` epochs.

    Args:
        engine (``BaseEngine``): Specifies the engine.
        freq (int): Specifies the frequency.
    """

    def __init__(self, engine: BaseEngine, freq: int):
        self._engine = engine
        self._freq = int(freq)

    def __call__(self) -> bool:
        r"""Evaluates the condition given the current state.

        Returns:
            bool: ``True`` if the condition is ``True``, otherwise
                ``False``.
        """
        return self._engine.epoch % self._freq == 0

    def __eq__(self, other) -> bool:
        if isinstance(other, EpochPeriodicCondition):
            return self.freq == other.freq
        return False

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(freq={self._freq:,}, epoch={self._engine.epoch:,})"

    @property
    def freq(self) -> int:
        r"""int: The frequency of the condition."""
        return self._freq


class IterationPeriodicCondition:
    r"""Implements an iteration periodic condition.

    This condition is true every ``freq`` iterations.

    Args:
        engine (``BaseEngine``): Specifies the engine.
        freq (int): Specifies the frequency.
    """

    def __init__(self, engine: BaseEngine, freq: int):
        self._engine = engine
        self._freq = int(freq)

    def __call__(self) -> bool:
        r"""Evaluates the condition given the current state.

        Returns:
            bool: ``True`` if the condition is ``True``, otherwise ``False``.
        """
        return self._engine.iteration % self._freq == 0

    def __eq__(self, other) -> bool:
        if isinstance(other, IterationPeriodicCondition):
            return self.freq == other.freq
        return False

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(freq={self._freq:,}, "
            f"iteration={self._engine.iteration:,})"
        )

    @property
    def freq(self) -> int:
        r"""int: The frequency of the condition."""
        return self._freq



r"""This module implements an asset manager."""

__all__ = ["AssetManager", "AssetNotFoundError"]

import logging
from typing import Any

logger = logging.getLogger(__name__)


class AssetManager:
    r"""Implements an asset manager."""

    def __init__(self):
        self._assets = {}

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def add_asset(self, name: str, asset: Any) -> None:
        r"""Adds an asset to the asset manager.

        Note that the name should be unique. If the name exists, the
        old asset will be overwritten by the new asset.

        Args:
            name (str): Specifies the name of the asset to add.
            asset: Specifies the asset to add.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils import AssetManager
            >>> manager = AssetManager()
            >>> manager.add_asset('mean', 5)
        """
        self._assets[name] = asset

    def get_asset(self, name: str) -> Any:
        r"""Gets an asset.

        Args:
            name (str): Specifies the asset to get.

        Returns:
            The asset

        Raises:
            ``AssetNotFoundError`` if the asset does not exist.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils import AssetManager
            >>> manager = AssetManager()
            >>> manager.add_asset('mean', 5)
            >>> manager.get_asset('mean')
            5
        """
        if name not in self._assets:
            raise AssetNotFoundError(f"The asset {name} does not exist")
        return self._assets[name]

    def has_asset(self, name: str) -> bool:
        r"""Indicates if the asset exists or not.

        Args:
            name (str): Specifies the name of the asset.

        Returns:
            bool: ``True`` if the asset exists, otherwise ``False``

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils import AssetManager
            >>> manager = AssetManager()
            >>> manager.has_asset('mean')
            False
            >>> manager.add_asset('mean', 5)
            >>> manager.has_asset('mean')
            True
        """
        return name in self._assets

    def remove_asset(self, name: str) -> None:
        r"""Removes an asset.

        Args:
            name (str): Specifies the name of the asset to remove.

        Raises:
            ``AssetNotFoundError`` if the asset does not exist.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils import AssetManager
            >>> manager = AssetManager()
            >>> manager.add_asset('mean', 5)
            >>> manager.remove_asset('mean')
            5
            >>> manager.has_asset('mean')
            False
        """
        if name not in self._assets:
            raise AssetNotFoundError(
                f"The asset {name} does not exist so it is not possible to remove it"
            )
        del self._assets[name]


class AssetNotFoundError(Exception):
    r"""Raised when trying to access an asset that does not exist."""



__all__ = ["NoRepeatRunner"]

from datetime import datetime
from pathlib import Path
from typing import Union

from gravitorch.runners.base import BaseRunner
from gravitorch.runners.utils import setup_runner
from gravitorch.utils.format import str_add_indent
from gravitorch.utils.io import save_text
from gravitorch.utils.path import sanitize_path


class NoRepeatRunner(BaseRunner):
    r"""Implements a runner that does not repeat a successful run.

    This runner logs if a run was successful. If a previous run was
    successful, this runner does not execute the logic again.

    Args:
        runner (``BaseRunner`` or dict): Specifies the runner or its
            configuration.
        path (``Path`` or str): Specifies the path where to log a
            successful run.
    """

    def __init__(self, runner: Union[Union[BaseRunner, dict]], path: Union[Path, str]):
        self._path = sanitize_path(path)
        self._success_path = self._path.joinpath("_MLTORCH_SUCCESS_")
        self._runner = setup_runner(runner)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  runner={str_add_indent(str(self._runner))},\n"
            f"  path={self._path},\n"
            ")"
        )

    def run(self) -> None:
        r"""Executes the logic of the runner if it was not successful
        before."""
        if not self._success_path.is_file():
            self._runner.run()
            save_text(str(datetime.now()), self._success_path)



r"""This module defines the normalized mean squared error metric."""

__all__ = ["NormalizedMeanSquaredError"]

import logging
from typing import Optional

from torch import Tensor
from torch.nn import functional as F

from gravitorch.distributed.ddp import SUM, sync_reduce
from gravitorch.engines.base import BaseEngine
from gravitorch.models.metrics.base import EmptyMetricError
from gravitorch.models.metrics.base_epoch import BaseEpochMetric
from gravitorch.utils.exp_trackers import EpochStep
from gravitorch.utils.format import str_scalar
from gravitorch.utils.history import MinScalarHistory

logger = logging.getLogger(__name__)


class NormalizedMeanSquaredError(BaseEpochMetric):
    r"""Implements the normalized mean squared error (NMSE) metric.

    Note: this metric does not work if all the targets are zero.

    Args:
        mode (str): Specifies the mode (e.g. train or eval).
        name (str, optional): Specifies the name of the metric.
            The name is used to log the metric results.
            Default: ``'nmse'``
    """

    def __init__(self, mode: str, name: str = "nmse"):
        super().__init__(mode=mode, name=name)
        self._sum_squared_errors = 0.0
        self._sum_squared_targets = 0.0
        self._num_predictions = 0
        self.reset()

    def extra_repr(self) -> str:
        return f"mode={self._mode}, name={self._name}"

    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches current metric to the provided engine.

        This method can be used to:

            - add event handler to the engine
            - set up history trackers

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        super().attach(engine)
        engine.add_history(MinScalarHistory(name=self._metric_name))

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the mean squared error metric given a mini-batch of
        examples.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the tensor of predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the tensor of targets.
        """
        self._sum_squared_errors += F.mse_loss(
            prediction.float(), target.float(), reduction="sum"
        ).item()
        self._sum_squared_targets += target.pow(2).sum().item()
        self._num_predictions += target.numel()

    def reset(self) -> None:
        r"""Resets the metric."""
        self._sum_squared_errors = 0.0
        self._sum_squared_targets = 0.0
        self._num_predictions = 0

    def value(self, engine: Optional[BaseEngine] = None) -> dict:
        r"""Evaluates the metric and log the results given all the examples
        previously seen.

        Args:
            engine (``BaseEngine``, optional): Specifies the engine.
                This argument is required to log the results in the
                engine. Default: ``None``.

        Returns:
             dict: The results of the metric.
        """
        num_predictions = sync_reduce(self._num_predictions, op=SUM)
        if not num_predictions:
            raise EmptyMetricError(f"{self.__class__.__qualname__} is empty")

        results = {
            self._metric_name: sync_reduce(self._sum_squared_errors, op=SUM)
            / sync_reduce(self._sum_squared_targets, op=SUM),
            f"{self._metric_name}_num_predictions": num_predictions,
        }
        for name, value in results.items():
            logger.info(f"{name}: {str_scalar(value)}")
        if engine:
            engine.log_metrics(results, EpochStep(engine.epoch))
        return results



r"""This module implements some utility functions for the parameter
initializers."""

__all__ = ["setup_parameter_initializer"]

import logging
from typing import Union

from gravitorch.utils.format import str_target_object
from gravitorch.utils.parameter_initializers.base import BaseParameterInitializer
from gravitorch.utils.parameter_initializers.no_init import NoParameterInitializer

logger = logging.getLogger(__name__)


def setup_parameter_initializer(
    parameter_initializer: Union[BaseParameterInitializer, dict, None]
) -> BaseParameterInitializer:
    r"""Sets up the model parameter initializer.

    Args:
        parameter_initializer (``BaseParameterInitializer`` or dict or
            ``None``): Specifies the model parameter initializer or
            its configuration. If ``None``, the
            ``NoParameterInitializer`` will be instantiated.

    Returns:
        ``BaseParameterInitializer``: The instantiated model parameter
            initializer.
    """
    if parameter_initializer is None:
        parameter_initializer = NoParameterInitializer()
    if isinstance(parameter_initializer, dict):
        logger.debug(
            "Initializing a parameter initializer from its configuration... "
            f"{str_target_object(parameter_initializer)}"
        )
        parameter_initializer = BaseParameterInitializer.factory(**parameter_initializer)
    return parameter_initializer



r"""This module defines the base model."""

__all__ = ["attach_module_to_engine", "setup_model", "setup_and_attach_model"]

import logging
from typing import Union

from torch.nn import Module

from gravitorch.engines.base import BaseEngine
from gravitorch.models.base import BaseModel
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def attach_module_to_engine(module: Module, engine: BaseEngine) -> None:
    r"""Attaches a module to the engine if the module has the ``attach`` method.

    This function does nothing if the module does not have a
    ``attach`` method.

    Args:
        module (``torch.nn.Module``): Specifies the module to attach
            to the engine.
        engine (``BaseEngine``): Specifies the engine.
    """
    if hasattr(module, "attach"):
        module.attach(engine)


def setup_model(model: Union[Module, dict]) -> Module:
    r"""Sets up the model.

    The model is instantiated from its configuration by using the
    ``BaseModel`` factory function.

    Args:
        model (``torch.nn.Module`` or dict): Specifies the model or
            its configuration.

    Returns:
        ``torch.nn.Module``: The (instantiated) model.
    """
    if isinstance(model, dict):
        logger.info(f"Initializing a model from its configuration... {str_target_object(model)}")
        model = BaseModel.factory(**model)
    return model


def setup_and_attach_model(engine: BaseEngine, model: Union[Module, dict]) -> Module:
    r"""Sets up and attaches the model to the engine.

    The model is attached to the engine by using the ``attach`` method.
     If the model does not have a ``attach`` method, the ``attach``
     step is skipped.

    Note that if you call this function ``N`` times with the same
    model, the model will be attached ``N`` times to the engine.

    Args:
        engine (``BaseEngine``): Specifies the engine.
        model (``torch.nn.Module`` or dict): Specifies the model or
            its configuration.

    Returns:
        ``torch.nn.Module``: The (instantiated) model.
    """
    model = setup_model(model)
    if hasattr(model, "attach"):
        logger.info("Adding a model to the engine state...")
        model.attach(engine=engine)
    return model



__all__ = ["VanillaModelCreator"]

import logging
from typing import Union

from torch import nn

from gravitorch import constants as ct
from gravitorch.creators.model.base import BaseModelCreator
from gravitorch.engines.base import BaseEngine
from gravitorch.models.utils import setup_and_attach_model, setup_model
from gravitorch.utils.device_placement import (
    AutoDevicePlacement,
    BaseDevicePlacement,
    setup_device_placement,
)
from gravitorch.utils.format import str_add_indent

logger = logging.getLogger(__name__)


class VanillaModelCreator(BaseModelCreator):
    r"""Implements a vanilla model creator.

    This model creator is designed for models that run on a single
    device. If ``device_placement=True``, the device is managed by the
    function ``gravitorch.distributed.device()``.

    Args:
        model_config (dict): Specifies the model configuration.
        attach_model_to_engine (bool, optional): If ``True``, the
            model is attached to the engine. Default: ``True``
        add_module_to_engine (bool, optional): If ``True``, the model
            is added to the engine state, so the model state is stored
            when the engine creates a checkpoint. Default: ``True``
        device_placement (bool, optional): Specifies the device
            placement module. This module moves the model on a target
            device. If ``None``, an ``AutoDevicePlacement`` object is
            instantiated. Default: ``None``
    """

    def __init__(
        self,
        model_config: dict,
        attach_model_to_engine: bool = True,
        add_module_to_engine: bool = True,
        device_placement: Union[BaseDevicePlacement, dict, None] = None,
    ):
        self._model_config = model_config
        self._attach_model_to_engine = bool(attach_model_to_engine)
        self._add_module_to_engine = bool(add_module_to_engine)
        self._device_placement = setup_device_placement(device_placement or AutoDevicePlacement())

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  attach_model_to_engine={self._attach_model_to_engine},\n"
            f"  add_module_to_engine={self._add_module_to_engine},\n"
            f"  device_placement={str_add_indent(self._device_placement)},\n"
            ")"
        )

    def create(self, engine: BaseEngine) -> nn.Module:
        logger.info("Creating model...")
        if self._attach_model_to_engine:
            model = setup_and_attach_model(engine=engine, model=self._model_config)
        else:
            model = setup_model(model=self._model_config)
        model = self._device_placement.send(model)
        if self._add_module_to_engine:
            logger.info(f"Adding a model to the engine state (key: {ct.MODEL})...")
            engine.add_module(ct.MODEL, model)
        return model



__all__ = [
    "AsinhMSELoss",
    "MSLELoss",
    "RelativeMSELoss",
    "RelativeSmoothL1Loss",
    "SymlogMSELoss",
    "SymmetricRelativeSmoothL1Loss",
]

from torch import Tensor
from torch.nn import Module

from gravitorch.nn.functional import (
    asinh_mse_loss,
    msle_loss,
    relative_mse_loss,
    relative_smooth_l1_loss,
    symlog_mse_loss,
    symmetric_relative_smooth_l1_loss,
)


class AsinhMSELoss(Module):
    r"""Implements the mean squared error (MSE) loss with inverse hyperbolic
    sine (arcsinh) transformation.

    Args:
        reduction (string, optional): Specifies the reduction to apply
            to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn import AsinhMSELoss
        >>> criterion = AsinhMSELoss()
        >>> loss = criterion(torch.randn(3, 5, requires_grad=True), torch.randn(3, 5))
        >>> loss.backward()
    """

    def __init__(self, reduction: str = "mean") -> None:
        super().__init__()
        self.reduction = str(reduction)

    def extra_repr(self) -> str:
        return f"reduction={self.reduction}"

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        return asinh_mse_loss(input, target, self.reduction)


class MSLELoss(Module):
    r"""Implements the mean squared logarithmic error (MSLE) loss.

    Note: this loss only works with positive values (0 included).

    Args:
        reduction (string, optional): Specifies the reduction to apply
            to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn import MSLELoss
        >>> criterion = MSLELoss()
        >>> loss = criterion(torch.rand(3, 5, requires_grad=True), torch.rand(3, 5))
        >>> loss.backward()
    """

    def __init__(self, reduction: str = "mean") -> None:
        super().__init__()
        self.reduction = str(reduction)

    def extra_repr(self) -> str:
        return f"reduction={self.reduction}"

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        return msle_loss(input, target, self.reduction)


class SymlogMSELoss(Module):
    r"""Implements the mean squared error (MSE) loss with symmetric logarithmic
    (symlog) transformation.

    Args:
        reduction (string, optional): Specifies the reduction to apply
            to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn import SymlogMSELoss
        >>> criterion = SymlogMSELoss()
        >>> loss = criterion(torch.randn(3, 5, requires_grad=True), torch.randn(3, 5))
        >>> loss.backward()
    """

    def __init__(self, reduction: str = "mean") -> None:
        super().__init__()
        self.reduction = str(reduction)

    def extra_repr(self) -> str:
        return f"reduction={self.reduction}"

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        return symlog_mse_loss(input, target, self.reduction)


class RelativeMSELoss(Module):
    r"""Implements the relative mean squared error.

    Args:
        reduction (string, optional): Specifies the reduction to apply
            to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``
        eps (float, optional): Specifies an arbitrary small strictly
            positive number to avoid undefined results when the target
            is zero. Default: ``1e-8``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn import RelativeMSELoss
        >>> criterion = RelativeMSELoss()
        >>> loss = criterion(torch.randn(3, 5, requires_grad=True), torch.randn(3, 5))
        >>> loss.backward()
    """

    def __init__(self, reduction: str = "mean", eps: float = 1e-8) -> None:
        super().__init__()
        self.reduction = str(reduction)
        self._eps = float(eps)

    def extra_repr(self) -> str:
        return f"reduction={self.reduction}"

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        return relative_mse_loss(input, target, reduction=self.reduction, eps=self._eps)


class RelativeSmoothL1Loss(Module):
    r"""Implements the relative smooth L1 loss.

    Args:
        reduction (string, optional): Specifies the reduction to apply
            to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``
        beta (float, optional): Specifies the threshold at which to
            change between L1 and L2 loss. The value must be
            non-negative. Default: ``1.0``
        eps (float, optional): Specifies an arbitrary small strictly
            positive number to avoid undefined results when the target
            is zero. Default: ``1e-8``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn import RelativeSmoothL1Loss
        >>> criterion = RelativeSmoothL1Loss()
        >>> loss = criterion(torch.randn(3, 5, requires_grad=True), torch.randn(3, 5))
        >>> loss.backward()
    """

    def __init__(self, reduction: str = "mean", beta: float = 1.0, eps: float = 1e-8):
        super().__init__()
        self.reduction = str(reduction)
        self._beta = float(beta)
        self._eps = float(eps)

    def extra_repr(self) -> str:
        return f"reduction={self.reduction}, beta={self._beta}, eps={self._eps}"

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        return relative_smooth_l1_loss(
            input,
            target,
            reduction=self.reduction,
            beta=self._beta,
            eps=self._eps,
        )


class SymmetricRelativeSmoothL1Loss(Module):
    r"""Implements the symmetric relative smooth L1 loss.

    Args:
        reduction (string, optional): Specifies the reduction to apply
            to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``
        beta (float, optional): Specifies the threshold at which to
            change between L1 and L2 loss. The value must be
            non-negative. Default: ``1.0``
        eps (float, optional): Specifies an arbitrary small strictly
            positive number to avoid undefined results when the target
            is zero. Default: ``1e-8``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn import SymmetricRelativeSmoothL1Loss
        >>> criterion = SymmetricRelativeSmoothL1Loss()
        >>> loss = criterion(torch.randn(3, 5, requires_grad=True), torch.randn(3, 5))
        >>> loss.backward()
    """

    def __init__(self, reduction: str = "mean", beta: float = 1.0, eps: float = 1e-8):
        super().__init__()
        self.reduction = str(reduction)
        self._beta = float(beta)
        self._eps = float(eps)

    def extra_repr(self) -> str:
        return f"reduction={self.reduction}, beta={self._beta}, eps={self._eps}"

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        return symmetric_relative_smooth_l1_loss(
            input,
            target,
            reduction=self.reduction,
            beta=self._beta,
            eps=self._eps,
        )



__all__ = ["PickleSaverIterDataPipe", "PyTorchSaverIterDataPipe"]

import logging
from collections.abc import Iterator
from pathlib import Path
from typing import Union

from torch.utils.data import IterDataPipe

from gravitorch.utils.format import str_add_indent
from gravitorch.utils.io import save_pickle, save_pytorch
from gravitorch.utils.path import sanitize_path

logger = logging.getLogger(__name__)


class PickleSaverIterDataPipe(IterDataPipe[Path]):
    r"""Implements a DataPipe to save each value from the source DataPipe in a
    pickle file.

    This DataPipe returns the path to each pickle file. If the source
    DataPipe has ``M`` values, ``M`` pickle files are created.

    Args:
        datapipe: Specifies the source DataPipe.
        root_path (``pathlib.Path`` or str): Specifies the directory
            where to save the pickle files.
        pattern (str, optional): Specifies the filename pattern of
            the file. The pattern should have ``'{index}'`` or similar
            syntax to indicate the index of the file.
            Default: ``'data_{index:04d}.pkl'``
    """

    def __init__(
        self,
        datapipe: IterDataPipe,
        root_path: Union[Path, str],
        pattern: str = "data_{index:04d}.pkl",
    ):
        self._datapipe = datapipe
        self._root_path = sanitize_path(root_path)
        if not self._root_path.is_dir():
            raise ValueError(f"root_path has to be a directory (received: {self._root_path})")
        self._pattern = str(pattern)
        if "index" not in self._pattern:
            raise ValueError(f"pattern does not have 'index' (received: {self._pattern})")

    def __iter__(self) -> Iterator[Path]:
        for i, data in enumerate(self._datapipe):
            path = self._root_path.joinpath(self._pattern.format(index=i))
            save_pickle(data, path)
            yield path

    def __len__(self) -> int:
        return len(self._datapipe)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  root_path={self._root_path},\n"
            f"  pattern={self._pattern},\n"
            f"  datapipe={str_add_indent(self._datapipe)},\n)"
        )


class PyTorchSaverIterDataPipe(IterDataPipe[Path]):
    r"""Implements a DataPipe to save each value from the source DataPipe in a
    PyTorch file.

    This DataPipe returns the path to each PyTorch file. If the source
    DataPipe has ``M`` values, ``M`` PyTorch files are created.

    Args:
        datapipe: Specifies the source DataPipe.
        root_path (``pathlib.Path`` or str): Specifies the directory
            where to save the PyTorch files.
        pattern (str, optional): Specifies the filename pattern of the
            file. The pattern should have ``'{index}'`` or similar
            syntax to indicate the index of the file.
            Default: ``'data_{index:04d}.pt'``
    """

    def __init__(
        self,
        datapipe: IterDataPipe,
        root_path: Union[Path, str],
        pattern: str = "data_{index:04d}.pt",
    ):
        self._datapipe = datapipe
        self._root_path = sanitize_path(root_path)
        if not self._root_path.is_dir():
            raise ValueError(f"root_path has to be a directory (received: {self._root_path})")
        self._pattern = str(pattern)
        if "index" not in self._pattern:
            raise ValueError(f"pattern does not have 'index' (received: {self._pattern})")

    def __iter__(self) -> Iterator[Path]:
        for i, data in enumerate(self._datapipe):
            path = self._root_path.joinpath(self._pattern.format(index=i))
            save_pytorch(data, path)
            yield path

    def __len__(self) -> int:
        return len(self._datapipe)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  root_path={self._root_path},\n"
            f"  pattern={self._pattern},\n"
            f"  datapipe={str_add_indent(self._datapipe)},\n)"
        )



# This constant is used to define the input of a model.
INPUT = "input"
# This constant is used to define the target to train/evaluate a model.
TARGET = "target"
# This constant is used to define the length of a sequence.
LENGTH = "length"
# This constant is used to define the index of an example in the dataset.
INDEX = "index"
# This constant is used to define the name of an example in the dataset.
NAME = "name"
# This constant is used to define the key of the runner in the config.
RUNNER = "runner"

# Some constants related to distributed training.
DISTRIBUTED = "distributed"
LAUNCHER = "launcher"
WORKER = "worker"

# These constants are used as default name for the training and evaluation metrics.
TRAIN = "train"
EVAL = "eval"

# This constant is used to define the default loss name. This key should be in the model output.
LOSS = "loss"
PREDICTION = "prediction"
MASK = "mask"
OUTPUT = "output"

# Engine keys
CHECKPOINT = "checkpoint"
DATA = "data"
DATA_SOURCE = "data_source"
ENGINE = "engine"
EVALUATION_LOOP = "evaluation_loop"
EXP_TRACKER = "exp_tracker"
LR_SCHEDULER = "lr_scheduler"
MODEL = "model"
OPTIMIZER = "optimizer"
STATE = "state"
TRAINING_LOOP = "training_loop"
SCALER = "scaler"

EARLY_STOPPING = "early_stopping"

ARTIFACT_FOLDER_NAME = "artifacts"
CHECKPOINT_FOLDER_NAME = "checkpoints"
